package.preload['projector.projector_proxy_notifies'] = (function (...)
--[[=============================================================================
    Projector Proxy Notification Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--[[
	Notify: INPUT_CHANGED
	Sent to the proxy to indicate the currently selected Input has changed
	Parameters
		bindingID - proxy id of proxy bound to input connection
		input - mod 1000 value of Input Connection ID
--]]
function NOTIFY.INPUT_CHANGED(bindingID, input)
    local tParams = {INPUT = input}
    SendNotify("INPUT_CHANGED", tParams, bindingID)
end
------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
    LogTrace "NOTIFY.ON()"
	SendSimpleNotify("ON")
end

function NOTIFY.OFF()
    LogTrace "NOTIFY.OFF()"
	SendSimpleNotify("OFF")
end
------------------------------ Volume Notification Code ------------------------------
--[[
	Notify: VOLUME_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Volume level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 level uses a percentage scale: 0 - 100
--]]
function NOTIFY.VOLUME_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("VOLUME_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BASS_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Bass level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BASS_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BASS_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: TREBLE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Treble level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.TREBLE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("TREBLE_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BALANCE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Balance level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BALANCE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BALANCE_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: LOUNDENSS_CHANGED
	Sent to the Projector proxy to indicate that the Loudness state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.LOUDNESS_CHANGED(bindingID, output, state)
    local tParams = {LOUDNESS = state, OUTPUT = output + 4000}
    SendNotify("LOUDNESS_CHANGED", tParams, bindingID)		
end

--[[
	Notify: MUTE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Mute state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.MUTE_CHANGED(bindingID, output, state)
    local tParams = {MUTE = state, OUTPUT = output + 4000}
    SendNotify("MUTE_CHANGED", tParams, bindingID)		
end
 end)
package.preload['projector.projector_proxy_commands'] = (function (...)
--[[=============================================================================
    ReceivedFromProxy Code for the Projector Proxy

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.ON(idBinding, tParams)
	gProjectorProxy:prx_ON(tParams)
end

function PRX_CMD.OFF(idBinding, tParams)
	gProjectorProxy:prx_OFF(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gProjectorProxy:prx_SET_INPUT(idBinding, tParams)
end

function PRX_CMD.TV_VIDEO(idBinding, tParams)
	gProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	gProjectorProxy:prx_MUTE_OFF(tParams)
end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	gProjectorProxy:prx_MUTE_ON(tParams)	
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	gProjectorProxy:prx_MUTE_TOGGLE(tParams)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	gProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_UP(tParams)
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_DOWN(tParams)
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_UP(tParams)
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_DOWN(tParams)
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_UP(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.INFO(idBinding, tParams)
	gProjectorProxy:prx_INFO(idBinding, tParams)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	gProjectorProxy:prx_GUIDE(idBinding, tParams)
end

function PRX_CMD.MENU(idBinding, tParams)
	gProjectorProxy:prx_MENU(idBinding, tParams)
end

function PRX_CMD.CANCEL(idBinding, tParams)
	gProjectorProxy:prx_CANCEL(idBinding, tParams)
end

function PRX_CMD.UP(idBinding, tParams)
	gProjectorProxy:prx_UP(idBinding, tParams)
end

function PRX_CMD.DOWN(idBinding, tParams)
	gProjectorProxy:prx_DOWN(idBinding, tParams)
end

function PRX_CMD.LEFT(idBinding, tParams)
	gProjectorProxy:prx_LEFT(idBinding, tParams)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	gProjectorProxy:prx_RIGHT(idBinding, tParams)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_DOWN(idBinding, tParams)
end

function PRX_CMD.START_UP(idBinding, tParams)
	gProjectorProxy:prx_START_UP(idBinding, tParams)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	gProjectorProxy:prx_START_LEFT(idBinding, tParams)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_START_RIGHT(idBinding, tParams)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_UP(idBinding, tParams)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	gProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
end

function PRX_CMD.ENTER(idBinding, tParams)
	gProjectorProxy:prx_ENTER(idBinding, tParams)
end

function PRX_CMD.RECALL(idBinding, tParams)
	gProjectorProxy:prx_RECALL(idBinding, tParams)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
	gProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.NUMBER_0(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_0(idBinding, tParams)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_1(idBinding, tParams)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_2(idBinding, tParams)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_3(idBinding, tParams)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_4(idBinding, tParams)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_5(idBinding, tParams)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_6(idBinding, tParams)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_7(idBinding, tParams)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_8(idBinding, tParams)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_9(idBinding, tParams)
end

function PRX_CMD.STAR(idBinding, tParams)
	gProjectorProxy:prx_STAR(idBinding, tParams)
end

function PRX_CMD.POUND(idBinding, tParams)
	gProjectorProxy:prx_POUND(idBinding, tParams)
end
 end)
package.preload['projector.projector_proxy_class'] = (function (...)
--[[=============================================================================
    Projector Proxy Class Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

DEFAULT_POWER_STATUS_QUERY_INTERVAL = 3         	-- Amount of time to wait BEFORE sending the next power status query, in seconds

ProjectorProxy = inheritsFrom(nil)

function ProjectorProxy:construct(projectorBindingID, bProcessesDeviceMessages, tVolumeRampingTracking)
	-- member variables
	self._ProjectorBindingID = projectorBindingID
	self._PowerState = ""							--Valid Values: "ON", "OFF", "POWER_ON_SEQUENCE", "POWER_OFF_SEQUENCE"
	self._VolumeIsRamping = false
	self._VolumeRamping = tVolumeRampingTracking	--{["state"] = false, ["mode"] = "",} ||	"state" is boolean, "mode" values: "VOLUME_UP" & "VOLUME_DOWN"
	self._LastVolumeStatusValue = 0	
	self._MenuNavigationInProgress = false
    self._MenuNavigationMode = ""
	self._MenuNavigationProxyID = ""	
	self._CurrentlySelectedInput = ""
	self._ProcessesDeviceMessages = bProcessesDeviceMessages
	self._ControlMethod = ""						--Valid Values: "NETWORK", "SERIAL", "IR" 
	self._PowerStatusQueryDelayS = DEFAULT_COMMAND_RESPONSE_INTERVAL
end

function ProjectorProxy:Initialize(PowerStatusQueryDelayInterval, CallbackParam)
	if (PowerStatusQueryDelayInterval ~= nil) then
		self._PowerStatusQueryDelayS = PowerStatusQueryDelayInterval
	end
	self._PowerStatusQueryDelayTimer = c4_timer:new("PowerStatusQueryDelay", self._PowerStatusQueryDelayS, "SECONDS", ProjectorProxy.OnPowerStatusQueryDelayTimerExpired, false, self)
end

-- Note the '.' instead of the ':'
function ProjectorProxy.OnPowerStatusQueryDelayTimerExpired(Instance)
	LogTrace("Power Status Query Timer expired")
	Instance._PowerStatusQueryDelayTimer:KillTimer()
	
	if (GetDevicePowerStatus ~= nil and type(GetDevicePowerStatus) == "function") then
		GetDevicePowerStatus(true)  --"true" sends command as a priority command
	end
end

------------------------------------------------------------------------
-- Projector Proxy Commands (PRX_CMD)
------------------------------------------------------------------------
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_ON(tParams)
    if (self._PowerState ~= nil) then
		if (self._PowerState == 'ON') or (self._PowerState == 'POWER_ON_SEQUENCE') then 
			LogInfo("Power State is '" .. self._PowerState .. "', exiting ON")
			return 
		end	
	end	
	if (self._ProcessesDeviceMessages == false) then
		self._PowerState = 'ON'
	else	
		self._PowerState = 'POWER_ON_SEQUENCE'
	end	
	ON()
end

function ProjectorProxy:prx_OFF(tParams)
    if (self._ProcessesDeviceMessages == false) then
	   self._PowerState = 'OFF'
    else	
	   self._PowerState = 'POWER_OFF_SEQUENCE'
    end	
    OFF()
  
    self._CurrentlySelectedInput = -1
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_SET_INPUT(idBinding, tParams)
	local input = tonumber(tParams["INPUT"] % 1000)
	self._CurrentlySelectedInput = input
	if (gControlMethod == "IR") then			
		NOTIFY.INPUT_CHANGED(self._ProjectorBindingID, input)				
	end 		

	SET_INPUT(idBinding, input)
end

function ProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "TV_VIDEO")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_MUTE_OFF(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "False")
	end 		
	MUTE_OFF()
end

function ProjectorProxy:prx_MUTE_ON(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "True")
	end 		
	MUTE_ON()
end

function ProjectorProxy:prx_MUTE_TOGGLE(tParams)
	MUTE_TOGGLE()
end

function ProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
	local c4VolumeLevel = tonumber(tParams['LEVEL'])
	SET_VOLUME_LEVEL(c4VolumeLevel)
end

function ProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
	PULSE_VOL_DOWN()
end

function ProjectorProxy:prx_PULSE_VOL_UP(tParams)
	PULSE_VOL_UP()
end

function ProjectorProxy:prx_START_VOL_DOWN(tParams)
	self:ChangeVolume("START_VOL_DOWN")
end

function ProjectorProxy:prx_START_VOL_UP(tParams)
	self:ChangeVolume("START_VOL_UP")
end

function ProjectorProxy:prx_STOP_VOL_DOWN(tParams)
	self:ChangeVolume("STOP_VOL_DOWN")
end

function ProjectorProxy:prx_STOP_VOL_UP(tParams)
	self:ChangeVolume("STOP_VOL_UP")
end

---------------------- Volume Helper Functions ----------------------
function ProjectorProxy:ChangeVolume(command_name)
	if (command_name == "STOP_VOL_UP") or (command_name == "STOP_VOL_DOWN") then
		self._VolumeIsRamping = false
		self._VolumeRamping.state = false
		self._VolumeRamping.mode = ""
	elseif (command_name == "START_VOL_UP") then 
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_UP" 
		PULSE_VOL_UP()	
	elseif (command_name == "START_VOL_DOWN") then 	
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_DOWN"	
		PULSE_VOL_DOWN()		
	else
		LogWarn(command_name .. " not handled in ChangeVolume()")
	end
end

function ProjectorProxy:ContinueVolumeRamping()
	local command
	if (gControlMethod == "IR") then   
		if (self._VolumeRamping.mode == "VOLUME_UP") then
			PULSE_VOL_UP()	
		elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
			PULSE_VOL_DOWN()	
		else
			LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		end	
	else
		local volume = self._LastVolumeStatusValue
		local deviceVolumeLevel = self:GetNextVolumeCurveValue(volume)
		if (deviceVolumeLevel ~= nil) then
			self._LastVolumeStatusValue = deviceVolumeLevel
			SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)                                 
		else
			LogWarn("ContinueVolumeRamping() next value is nil")
			return
		end
	end 
end

function ProjectorProxy:GetNextVolumeCurveValue(volume)
	local i, point
	volume=tonumber(volume)
	if (self._VolumeRamping.mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		LogWarn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	  local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --LogInfo("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --LogInfo("vl = " .. vl)
	  vl= tonumber(("%.".."0".."f"):format(vl))
	  --LogInfo("vl new = " .. vl)
	  LogInfo("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_PULSE_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_START_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_START_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_PRESET_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_UP")
end

function ProjectorProxy:prx_PRESET_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_DOWN")
end

function ProjectorProxy:prx_SET_CHANNEL(idBinding, tParams)
    local channel= tParams['CHANNEL']
	SET_CHANNEL(idBinding, channel)
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_INFO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "INFO")
end

function ProjectorProxy:prx_GUIDE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "GUIDE")
end

function ProjectorProxy:prx_MENU(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "MENU")
end

function ProjectorProxy:prx_CANCEL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "CANCEL")
end

function ProjectorProxy:prx_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "UP")
end

function ProjectorProxy:prx_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "DOWN")
end

function ProjectorProxy:prx_LEFT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "LEFT")
end

function ProjectorProxy:prx_RIGHT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RIGHT")
end

function ProjectorProxy:prx_START_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_DOWN")
end

function ProjectorProxy:prx_START_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_UP")
end

function ProjectorProxy:prx_START_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_LEFT")
end

function ProjectorProxy:prx_START_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_RIGHT")
end

function ProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_DOWN")
end

function ProjectorProxy:prx_STOP_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_UP")
end

function ProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_LEFT")
end

function ProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_RIGHT")
end

function ProjectorProxy:prx_ENTER(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "ENTER")
end

function ProjectorProxy:prx_RECALL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RECALL")
end

function ProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "OPEN_CLOSE")
end

function ProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_A")
end

function ProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_B")
end

function ProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_C")
end

function ProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_D")
end

---------------------- Menu Navigation Helper Functions ----------------------
function ProjectorProxy:NavigateMenu(idBinding, command_name)
	if (command_name == "STOP_UP") 
				or (command_name == "STOP_DOWN") 
				or (command_name == "STOP_LEFT") 
				or (command_name == "STOP_RIGHT") then
		self._MenuNavigationInProgress = false
		self._MenuNavigationMode = ""
		self._MenuNavigationProxyID = ""
		return
	elseif (command_name == "START_UP") then 
		self._MenuNavigationMode = "UP"
	elseif (command_name == "START_DOWN") then 	
		self._MenuNavigationMode = "DOWN"	
	elseif (command_name == "START_LEFT") then 
		self._MenuNavigationMode = "LEFT"	
     elseif (command_name == "START_RIGHT") then 
		self._MenuNavigationMode = "RIGHT"
	else
		LogWarn(command_name .. " not handled in NavigateMenu()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, self._MenuNavigationMode)
	self._MenuNavigationInProgress = true
	self._MenuNavigationProxyID = idBinding
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_NUMBER_0(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_0")
end

function ProjectorProxy:prx_NUMBER_1(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_1")
end

function ProjectorProxy:prx_NUMBER_2(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_2")
end

function ProjectorProxy:prx_NUMBER_3(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_3")
end

function ProjectorProxy:prx_NUMBER_4(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_4")
end

function ProjectorProxy:prx_NUMBER_5(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_5")
end

function ProjectorProxy:prx_NUMBER_6(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_6")
end

function ProjectorProxy:prx_NUMBER_7(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_7")
end

function ProjectorProxy:prx_NUMBER_8(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_8")
end

function ProjectorProxy:prx_NUMBER_9(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_9")
end

function ProjectorProxy:prx_STAR(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "STAR")
end

function ProjectorProxy:prx_POUND(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "POUND")
end

------------------------------------------------------------------------
-- Projector Proxy Notifies
------------------------------------------------------------------------

function ProjectorProxy:dev_InputChanged(bindingID, input)
	NOTIFY.INPUT_CHANGED(bindingID, input)
end

function ProjectorProxy:dev_PowerOn()
	self._PowerState = "ON"
	NOTIFY.ON()	
end

function ProjectorProxy:dev_PowerOff()
	self._PowerState = "OFF"
	NOTIFY.OFF()
end

function ProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	NOTIFY.VOLUME_LEVEL_CHANGED(self._ProjectorBindingID, output, c4Level)	
	
	if (self._VolumeIsRamping) then
		--do nothing
		--during volume ramping, LastVolumeStatusValue is set in ContinueVolumeRamping()
	else
		self._LastVolumeStatusValue = deviceLevel
	end	
end

function ProjectorProxy:dev_MuteChanged(output, state)
	NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, output, state)
end		

function ProjectorProxy:dev_BassLevelChanged(output, level)
	NOTIFY.BASS_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_TrebleLevelChanged(output, level)
	NOTIFY.TREBLE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_BalanceLevelChanged(output, level)
	NOTIFY.BALANCE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_LoudnessChanged(output, state)
	NOTIFY.LOUDNESS_CHANGED(self._ProjectorBindingID, output, state)
end

function ProjectorProxy:dev_ChannelStatus(channel)
	NOTIFY.CHANNEL_CHANGED(self._ProjectorBindingID, channel)
end end)
package.preload['common.c4_url_connection'] = (function (...)
--[[=============================================================================
    Base for a url connection driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_url_connection = "2015.03.02"
end

UrlConnectionBase = inheritsFrom(DeviceConnectionBase)

function UrlConnectionBase:construct(Url)
	self.superClass():construct()
	self._Url = Url
end

function UrlConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "URL"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	OnURLConnectionChanged()
end

function UrlConnectionBase:ControlMethod()
	return "URL"
end

function UrlConnectionBase:SetUrl(Url)
	self._Url = Url
end

function UrlConnectionBase:SendCommand(sCommand, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(self._Url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(self._Url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:SendCommandUrl(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:ReceivedAsync(ticketId, sData, responseCode, tHeaders)
	local tMessage = {
		["ticketId"] = ticketId,
		["sData"] = sData,
		["responseCode"] = responseCode,
		["tHeaders"] = tHeaders
	}
	
	status, err = pcall(HandleMessage, tMessage)
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

function ConnectURL()
	gIsUrlConnected = true
	SetControlMethod()
end

function DisconnectURL()
	gIsUrlConnected = false
	SetControlMethod()
end
 end)
package.preload['connections'] = (function (...)
--[[=============================================================================
    Functions for managing the status of the drivers bindings and connection state
 
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_url_connection"

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.connections = "2015.03.31"
end

-- constants
COM_USE_ACK = false
COM_COMMAND_DELAY_MILLISECONDS = 250
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4

NETWORK_PORT = 20554
IR_BINDING_ID = 2
SERIAL_BINDING_ID = 1
NETWORK_BINDING_ID = 6000

--[[=============================================================================
    OnSerialConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when a serial binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnSerialConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnIRConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when an IR binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnIRConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnNetworkConnectionChanged(idBinding, bIsBound)
  
    Description:
    Function called when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkConnectionChanged(idBinding, bIsBound)
	
end

--[[=============================================================================
    OnNetworkStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Called when the network connection status changes. Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnNetworkStatusChanged(idBinding, nPort, sStatus)
	
end

--[[=============================================================================
    OnURLConnectionChanged(url)
  
    Description:
    Function called when the c4_url_connection is created.
  
    Parameters:
    url - url used by the url connection.
  
    Returns:
    None
===============================================================================]]
function OnURLConnectionChanged(url)
	
end

--[[=============================================================================
    DoEvents()
  
    Description:
    Called from OnSendTimeExpired in the DeviceConnectionBase.
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function DoEvents()
	if (gProjectorProxy._VolumeIsRamping == true) then
		if (gProjectorProxy._VolumeRamping.state == true) then gProjectorProxy:ContinueVolumeRamping() end
	end
end

--[[=============================================================================
    SendKeepAlivePollingCommand()
  
    Description:
    Sends a driver specific polling command to the connected system
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function SendKeepAlivePollingCommand()
    LogTrace("SendKeepAlivePollingCommand()")

    local command = "PW" --"Poll using power query"
    PackAndQueueReferenceCommand("SendKeepAlivePollingCommand", command)	
    
end
 end)
package.preload['proxy_commands'] = (function (...)
--[[=============================================================================
    Commands received from the Projector proxy (ReceivedFromProxy)

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: ON
	Parameters:
--]]
function ON()
    gPowerEventStarted = os.time()
    gLastPowerEvent = "POWER_ON"
    
    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
    
    gPowerOnTimer:StartTimer(command_delay, delay_units)
    
    local command
    if (gControlMethod == "IR") then
	   -- TODO: create packet/command to send to the device
	   command = CMDS_IR["ON"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("ON", command, command_delay, delay_units)
    else
	   if (gPowerOffTimer:TimerStarted()) then
		  gPowerOnDelayTimer:StartTimer(120, "SECONDS")	--tonumber(Properties["Power Off Delay Seconds"])
	   else
		  command = CMDS["ON"]
		  LogTrace("command = " .. command)
		  PackAndQueueCommand("ON", command, command_delay, delay_units)
		
		  gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
	   end
    end
end

function OnPowerQueryDelayTimerExpired()
	LogTrace("Power Query Delay Timer expired")
	gPowerQueryDelayTimer:KillTimer()
	
	GetDevicePowerStatus()
end

function OnPowerOnDelayTimerExpired()
    LogTrace("Power On Delay Timer expired")
    gPowerOnDelayTimer:KillTimer()

    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
    local command = CMDS["ON"]
    PackAndQueueCommand("ON", command, command_delay, delay_units)
    gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
end

function OnPowerOffDelayTimerExpired()
    LogTrace("Power Off Delay Timer expired")
    gPowerOffDelayTimer:KillTimer()

    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    local command = CMDS["OFF"]
    PackAndQueueCommand("OFF", command, command_delay, delay_units)	
    gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
end

function OnPowerOnTimerExpired()
    LogTrace("Power On Timer expired")
    gPowerOnTimer:KillTimer()
end

function OnPowerOffTimerExpired()
    LogTrace("Power Off Timer expired")
    gPowerOffTimer:KillTimer()
end


--[[
	Proxy Command: OFF
	Parameters:
--]]
function OFF()
    gPowerEventStarted = os.time()
    gLastPowerEvent = "POWER_OFF"
    
    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    
    gPowerOffTimer:StartTimer(command_delay, delay_units)
    
    local command 
    if (gControlMethod == "IR") then
		command = CMDS_IR["OFF"]
		LogTrace("command = " .. command)
		PackAndQueueCommand("OFF", command, command_delay, delay_units)		
    else
	   if (gPowerOnTimer:TimerStarted()) then
		  --hanlding a quick power off after a power on...needs a lot of time...
		  gPowerOffDelayTimer:StartTimer(120, "SECONDS")
	   else
		  command = CMDS["OFF"]
		  LogTrace("command = " .. command)
		  PackAndQueueCommand("OFF", command, command_delay, delay_units)	

		  gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
	   end
    end  
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: SET_INPUT
	Parameters:
		idBinding: proxybindingid of proxy bound to input connection
		output: mod 1000 value of Output Connection id	
		input: mod 1000 value of Input Connection id
--]]
function SET_INPUT(idBinding, input)
	local command
	if (gControlMethod == "IR") then		
	   command = tInputCommandMap_IR[tInputConnMapByID[input]]	
	   PackAndQueueCommand("SET_INPUT", command, 15, "SECONDS")
	else
	   command = tInputCommandMap[tInputConnMapByID[input]] 
	   local delay, delay_units
	   if (gProjectorProxy._PowerState == "POWER_ON_SEQUENCE") then
		  delay = 120
		  delay_units = "SECONDS"
	   else
		  delay = tonumber(Properties["Command Delay Milliseconds"])
		  delay_units = "MILLISECONDS"
	   end
	   gSetInputDelayTimer:StartTimer(delay, delay_units)
	   gSetInputDelayTimer._CallbackParam = command
	end 		
end

function OnSetInputDelayTimerExpired(command)
    LogTrace("OnSetInputDelayTimerExpired, command = " .. command)
    gSetInputDelayTimer:KillTimer()
    PackAndQueueCommand("SET_INPUT", command, 15, "SECONDS")
    PackAndQueueReferenceCommand("INPUT_QUERY", "IP")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: MUTE_OFF
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_OFF()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["MUTE_OFF"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_OFF"]
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_OFF", command)
end

--[[
	Proxy Command: MUTE_ON
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_ON()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_ON"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_ON"]
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_ON", command)		
end

--[[
	Proxy Command: MUTE_TOGGLE
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_TOGGLE()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_TOGGLE"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_TOGGLE"]
	end 
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_TOGGLE", command)
end

--[[
	Proxy Command: SET_VOLUME_LEVEL
	Parameters:
		output: mod 1000 value of Output Connection id	
		c4VolumeLevel: volume level to be set represented in Control4 scale (0-100)
--]]
function SET_VOLUME_LEVEL(c4VolumeLevel)
	-- TODO: edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local minDeviceLevel = -80
	local maxDeviceLevel = 20
	local deviceVolumeLevel = ConvertVolumeToDevice(c4VolumeLevel, minDeviceLevel, maxDeviceLevel)	
	
	---- TODO: uncomment and edit string padding, if required, based upon the protocol specification
    --deviceVolumeLevel = string.rep("0", 2 - string.len(deviceVolumeLevel)) .. deviceVolumeLevel
	
	LogInfo('deviceVolumeLevel: ' .. deviceVolumeLevel)
  
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	LogTrace("command = " .. command)
	PackAndQueueCommand("SET_VOLUME_LEVEL", command)
end

--[[
	Helper Function: SET_VOLUME_LEVEL_DEVICE
	Parameters:
		output: mod 1000 value of Output Connection id	
		deviceVolumeLevel: volume level to be set represented in device scale (as sepcified in the device's control protocol)
--]]
function SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)
	--Called from ContinueVolumeRamping()
	
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("SET_VOLUME_LEVEL_DEVICE", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_DOWN
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_DOWN()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_DOWN"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["PULSE_VOL_DOWN"]
	end 		
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_DOWN", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_UP
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_UP()
	local command
	if (gControlMethod == "IR") then	
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_UP"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["PULSE_VOL_UP"]
	end 		
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_UP", command, command_delay)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	Helper Function: SEND_COMMAND_FROM_COMMAND_TABLE
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be sent
--]]
function SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, command_name)
    LogTrace("SEND_COMMAND_FROM_COMMAND_TABLE(), idBinding=" .. idBinding .. ", command_name=" .. command_name)
	
	-- TODO: create packet/command to send to the device
	local command = GetCommandFromCommandTable(idBinding, command_name)
	
	if (command == nil) then
		LogTrace("command is nil")
	else
		LogTrace("command = " .. command)
		PackAndQueueCommand(command_name, command)
	end			
end

--[[
	Helper Function: GetCommandFromCommandTable
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be returned
--]]
function GetCommandFromCommandTable(idBinding, command_name)
	LogTrace("GetCommand()")
	local t = {}
	
	-- TODO: select/create appropriate commands table structure
	
	if (gControlMethod == "IR") then
		t = CMDS_IR
	else
		t = CMDS
	end	

	if (t[idBinding][command_name] ~= nil) then
		return t[idBinding][command_name]
	elseif (t[command_name] ~= nil) then
		return t[command_name]
	else
		LogWarn('GetCommandFromCommandTable: command not defined - '.. command_name)
		return nil
	end	
end

--[[
	Helper Function: GetDeviceVolumeStatus
--]]
function GetDeviceVolumeStatus()
    LogTrace("GetDeviceVolumeStatus()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_VOLUME_STATUS"]
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Volume", command)
	
	-- TODO: verify table entry in CMDS table
	command = CMDS["GET_MUTE_STATUS"]
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Mute", command)	
end

--[[
	Helper Function: GetLampHours
--]]
function GetLampHours()
    LogTrace("GetLampHours()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_LAMP_HOURS"]
	LogTrace("command = " .. command)
	PackAndQueueReferenceCommand("GetLampHours", command)
end

--[[
	Helper Function: GetDevicePowerStatus
--]]
function GetDevicePowerStatus(...)
	local bPriority = select(1, ...) or false
    LogTrace("GetDevicePowerStatus()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_POWER_STATUS"]
	LogTrace("command = " .. command)
	if (bPriority) then
		PackAndQueuePriority1ReferenceCommand("GetDevicePowerStatus: Volume", command)
	else
		PackAndQueueReferenceCommand("GetDevicePowerStatus: Volume", command)
	end
end

--[[
	Helper Function: GetNetworkSettings
--]]
function GetNetworkSettings()
    LogTrace("GetNetworkSettings()")
    PackAndQueueReferenceCommand("GetCurrentPort", "LSPT")
    
end

function SetNetworkPort(nPort)
    --Ex: Reset Port to 20555 (0x504B)
    local hPort = string.format("%X", nPort)
    PackAndQueueCommand("SetPortNumber", "LSPT" .. hPort)
    
    --Reboot Network
    --PackAndQueueCommand("RebootNetwork", "LSRS")
end
	 end)
package.preload['properties'] = (function (...)
--[[=============================================================================
    Properties Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

function ON_PROPERTY_CHANGED.NetworkKeepAliveIntervalSeconds(propertyValue)
	gNetworkKeepAliveInterval = tonumber(propertyValue)

end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)

	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end

 end)
package.preload['projector_init'] = (function (...)
--[[=============================================================================
    PROJECTOR Protocol Initialization Functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end


PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_EARLY_INIT.projector_init()

end

function ON_DRIVER_INIT.projector_init()
	--LogTrace("ON_DRIVER_INIT.ProtocolDeclarations()")
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
end

function ON_DRIVER_LATEINIT.projector_init()
end

function PROTOCOL_DECLARATIONS.SetGlobals()
	LogTrace("PROTOCOL_DECLARATIONS.SetGlobals()")

	--[[
	TODO: Determine projector operating conditions to be exposed for Composer Programming.
		Most projectors will report status, i.e. Warm Up,Cool Down and Error, etc.
		Lamp usage may also be reported. The sample variables below are referenced in the
		DEV_MSG.POWER() function in the device_messages.lua file.
	--]]
	C4:AddVariable('Projector_Operating_Status',0,'STRING',false,false)	
	--C4:AddVariable('Lamp_Hours',0,'STRING',false,false)
	gProjector_Operating_Status = ""
	gLamp_Hours = ""
end
	
function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--index:  Proxy Command Name
		--value:  IR Code ID	 
		--Power
		["ON"]             = "",
		["OFF"]            = "",
		
		--Input Toggle
		["INPUT_TOGGLE"]="",
		["TV_VIDEO"] 		= "",
		
		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",	
		["RECALL"]         = "",
		["PREV"]           = "",
		
		--Digits
		["NUMBER_0"]       = "", 	
		["NUMBER_1"]       = "",		
		["NUMBER_2"]       = "",	
		["NUMBER_3"]       = "",	
		["NUMBER_4"]       = "",	
		["NUMBER_5"]       = "",	
		["NUMBER_6"]       = "",	
		["NUMBER_7"]       = "",	
		["NUMBER_8"]       = "",	
		["NUMBER_9"]       = "",
		["STAR"]           = "",
		["DOT"]            = "",
		
		--Volume
		["MUTE_ON"]        = "",
		["MUTE_OFF"]       = "",
		["MUTE_TOGGLE"]    = "",		
		["PULSE_VOL_DOWN"] = "",
		["PULSE_VOL_UP"]   = "",		
		["SET_VOLUME_LEVEL"] = "",		
		
	}			
	
	CMDS_IR[PROJECTOR_PROXY_BINDINGID] = {}	
	
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()")

	CMDS = {
		['CANCEL']             = tohex ('52 43 37 33 30 33'), --back
		['CLOSED_CAPTIONED']   = '',
		['DOWN']               = tohex ('52 43 37 33 30 32'), 
		['ENTER']              = tohex ('52 43 37 33 32 46'), --OK
		['GUIDE']              = '',
		['HYPHEN']             = '',
		['INFO']               = '',
		['INPUT_TOGGLE']       = '',
		['LEFT']               = tohex ('52 43 37 33 33 36'),
		['LOUDNESS_OFF']       = '',
		['LOUDNESS_ON']        = '',
		['LOUDNESS_TOGGLE']    = '',
		['MENU']               = tohex ('52 43 37 33 32 45'),
		['MUTE_OFF']           = '',
		['MUTE_ON']            = '',
		['MUTE_TOGGLE']        = '',
		['NUMBER_0']           = '',
		['NUMBER_1']           = '',
		['NUMBER_2']           = '',
		['NUMBER_3']           = '',
		['NUMBER_4']           = '',
		['NUMBER_5']           = '',
		['NUMBER_6']           = '',
		['NUMBER_7']           = '',
		['NUMBER_8']           = '',
		['NUMBER_9']           = '',
		['OFF']                = tohex ('50 57 30'), 
		['ON']                 = tohex ('50 57 31'),
		['PAGE_DOWN']          = '',
		['PAGE_UP']            = '',
		['POUND']              = '',
		['PROGRAM_A']          = '',
		['PROGRAM_B']          = '',
		['PROGRAM_C']          = '',
		['PROGRAM_D']          = '',
		['PULSE_BALANCE_DOWN'] = '',
		['PULSE_BALANCE_UP']   = '',
		['PULSE_BASE_UP']      = '',
		['PULSE_BASS_DOWN']    = '',
		['PULSE_CH_DOWN']      = '',
		['PULSE_CH_UP']        = '',
		['PULSE_TREBLE_DOWN']  = '',
		['PULSE_TREBLE_UP']    = '',
		['PULSE_VOL_UP']       = '',
		['PUSLE_VOL_DOWN']     = '',
		['RECALL']             = tohex ('52 43 37 33 30 33'), --back
		['RIGHT']              = tohex ('52 43 37 33 33 34'),
		['SET_BALANCE']        = '',
		['SET_BASSS_LEVEL']    = '',
		['SET_CHANNEL']        = '',
		['SET_INPUT']          = '',
		['SET_TREBLE_LEVEL']   = '',
		['SET_VOLUME_LEVEL']   = '',
		['STAR']               = '',
		['START_BALANCE_DOWN'] = '',
		['START_BALANCE_UP']   = '',
		['START_BASS_DOWN']    = '',
		['START_BASS_UP']      = '',
		['START_CH_DOWN']      = '',
		['START_CH_UP']        = '',
		['START_TREBLE_DOWN']  = '',
		['START_TREBLE_UP']    = '',
		['START_VOL_DOWN']     = '',
		['START_VOL_UP']       = '',
		['STOP_BALANCE_DOWN']  = '',
		['STOP_BALANCE_UP']    = '',
		['STOP_BASS_DOWN']     = '',
		['STOP_BASS_UP']       = '',
		['STOP_CH_DOWN']       = '',
		['STOP_CH_UP']         = '',
		['STOP_TREBLE_DOWN']   = '',
		['STOP_TREBLE_UP']     = '',
		['STOP_VOL_DOWN']      = '',
		['STOP_VOL_UP']        = '',
		['TV_VIDEO']           = '',
		['UP']                 = tohex ('52 43 37 33 30 31'),
		['ANAMORPHIC_OFF']   = tohex ('49 4E 56 53 30'),
		['Trigger Off']        = tohex ('46 55 54 52 30'),
		['Trigger Power']   = tohex ('46 55 54 52 31'), 
		['Trigger Anamorphic']   = tohex ('46 55 54 52 32'),
		['STANDARD']   = tohex ('50 4D 50 52 30 33 0A'),
		['ANAMORPHIC_A']   = tohex ('49 4E 56 53 31'),
		['ANAMORPHIC_B']   = tohex ('49 4E 56 53 32'),
		['ANAMORPHIC_C']   = tohex ('49 4E 56 53 33'),
		['FILM'] = tohex ('50 4D 50 4D 30 30'),
		['CINEMA'] = tohex ('50 4D 50 4D 30 31'),
		['ANIMATION_1'] = tohex ('50 4D 50 52 30 36'),
		['ANIMATION_2'] = tohex ('50 4D 50 52 30 37'),
		['NATURAL'] = tohex ('50 4D 50 4D 30 33'),
		['STAGE'] = tohex ('50 4D 50 4D 30 34'),
		['THX'] = tohex ('50 4D 50 4D 30 36'),
		['3D'] = tohex ('50 4D 50 4D 30 42'),
		['Toggle'] = tohex ('52 43 37 33 44 37'), --User
		['1'] = tohex ('50 4D 50 4D 30 43'), --User
		['2'] = tohex ('50 4D 50 4D 30 44'), --User
		['3'] = tohex ('50 4D 50 4D 30 45'), --User
		['4'] = tohex ('50 4D 50 4D 30 46'), --User
		['5'] = tohex ('50 4D 50 4D 30 30'), --User
		['3D Format Off'] = tohex ('49 53 33 44 30'),
		['3D Format Auto'] = tohex ('49 53 33 44 31'),
		['3D Format Frame Packing'] = tohex ('49 53 33 44 32'),
		['3D Frame Side by Side'] = tohex ('49 53 33 44 33'),
		['3D Frame Top and Bottom'] = tohex ('49 53 33 44 34'),
		['2D to 3D Conversion Off'] = tohex ('49 53 33 43 30'),
		['2D to 3D Conversion On'] = tohex ('49 53 33 43 31'),
		['3D Format'] = tohex ('52 43 37 33 44 36'),
		['Hide Toggle'] = tohex ('52 43 37 33 31 44'), 
		['Hide On'] = tohex ('52 43 37 33 44 30'), 
		['Hide Off'] = tohex ('52 43 37 33 44 31'), 
		['ISF Day'] = tohex ('52 43 37 33 36 34'), 
		['ISF Night'] = tohex ('52 43 37 33 36 35'), 
		['ISF Off'] = tohex ('52 43 37 33 35 41'), 
		['ISF On'] = tohex ('52 43 37 33 35 42'), 
		['THX Bright'] = tohex ('52 43 37 33 38 35'), 
		['THX Dark'] = tohex ('52 43 37 33 38 36'), 
		['THX Off'] = tohex ('52 43 37 33 43 37'), 
		['THX On'] = tohex ('52 43 37 33 43 38'), 
		['Gamma A'] = tohex ('52 43 37 33 33 39'),
		['Gamma B'] = tohex ('52 43 37 33 33 41'),
		['Gamma C'] = tohex ('52 43 37 33 33 42'),
		['Gamma D'] = tohex ('52 43 37 33 33 46'),
		['Gamma 1'] = tohex ('52 43 37 33 33 43'),
		['Gamma 2'] = tohex ('52 43 37 33 33 44'),
		['Gamma 3'] = tohex ('52 43 37 33 33 45'),
		['Gamma Normal'] = tohex ('52 43 37 33 33 38'),
		['Gamma Cycle'] = tohex ('52 43 37 33 37 35'),
		['CMD Off'] = tohex ('50 4D 43 4D 30'), 
		['CMD Low'] = tohex ('50 4D 43 4D 33'), 
		['CMD High'] = tohex ('50 4D 43 4D 34'), 
		['CMD Inverse Telecine'] = tohex ('50 4D 43 4D 35'), 
		['Cycle CMD'] = tohex ('52 43 37 33 38 41'), 
		['16x9'] = tohex ('52 43 37 33 32 36'), --Aspect
		['4x3'] = tohex ('52 43 37 33 32 35'), --Aspect
		['Cycle'] = tohex ('52 43 37 33 37 37'), --Aspect
		['Lens Select 1'] = tohex ('49 4E 4D 4C 30'),
		['Lens Select 2'] = tohex ('49 4E 4D 4C 31'),
		['Lens Select 3'] = tohex ('49 4E 4D 4C 32'),
		['GET_POWER_STATUS'] = tohex ('50 57'),
	}


	CMDS['PROGRAM_A'] = CMDS[Properties["Red Button"]]
	CMDS['PROGRAM_B'] = CMDS[Properties["Green Button"]]
	CMDS['PROGRAM_C'] = CMDS[Properties["Yellow Button"]]
	CMDS['PROGRAM_D'] = CMDS[Properties["Blue Button"]]
	CMDS['STAR'] = CMDS[Properties["Star Button"]]
	CMDS['POUND'] = CMDS[Properties["Pound Button"]]
	
	CMDS[PROJECTOR_PROXY_BINDINGID] = {}
	
end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	LogTrace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	tInputCommandMap = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		["HDMI 1"] = tohex ('49 50 36'),
		["HDMI 2"] = tohex ('49 50 37'),
		["COMPONENT"] = tohex ('49 50 32'),
		["PC"] = tohex ('49 50 33'),
	}
	
	--tInputResponseMap = ReverseTable(tInputCommandMap)	-- Reverses the tInputCommandMap table	
	tInputResponseMap = {
		--index:  mod 1000 value of Input Connection id
		[6] = "HDMI 1",
		[7] = "HDMI 2", 	
		[2] = "COMPONENT",
		[3] = "PC",
	}	
	
	tInputCommandMap_IR = { 
		--index:  Connection Name
		--value:  IR Code ID
		--["ADAPTER PORT"] = "51362",
		[""] = "",	
	}

	----------------------------------------- [*I/O HELPER TABLES*] -----------------------------------------
	tOutputConnMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Output Connection Name
		[0] = "Main Output",
	}

	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		[0] = "HDMI 1",
		[1] = "HDMI 2", 	
		[2] = "COMPONENT",
		[3] = "PC",
	}
	
	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--["INPUT HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 2"] = {ID = 1,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["COMPONENT"] = {ID = 2,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["PC"] = {ID = 3,BindingID = PROJECTOR_PROXY_BINDINGID,},
	}

end	

function PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()")
	
end

function PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()")
	
	--[[
	TODO: Update tVolumeCurve table based upon devices volume range and optimal ramping steps
		tVolumeCurve is used to plot values that will be used
		to build volume commands during volume ramping.
		Specifically, they are used in GetNextVolumeCurveValue() which is 
		called from the ContinueVolumeRamping() function. 
		The values entered here should correlate to the volume parameters as defined in the device protocol.
		Below is an example, the table for your device may use different values.
		Also the number and size of steps may differ for your device. Adjust values to get 
		a smooth volume ramping from low to high volume.
		NOTE: values must be in decimal (i.e. not hexadecimal...)
	--]]
	tVolumeCurve = {
		-- 5 unit STEPS [0 to 20]
		0,5,10,15,20,
		-- 4 unit STEPS [24 to 58]
		24,28,32,36,38,
		42,46,49,52,55,58,
		-- 2 unit STEPS [60 to 90]
		60,62,64,66,68,
		70,72,74,76,78,
		80,82,84,86,88,90,           
		-- 1 unit STEPS [90 to 100] 
		91,92,93,94,95,
		96,97,98,99,100,                             
	}
	
end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

 end)
package.preload['device_messages'] = (function (...)
--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
	local message, pos
	--local cmdType = "[" .. string.pack('b', 0x06) .. string.pack('b', 0x40) .. "]"
	local stx = string.pack('bbb', 0x40, 0x89, 0x01)
	local etx = string.pack('b',0x0A)
	--local pattern = cmdType .. stx .. "(.-)" .. etx .. "()"
	local pattern = stx .. "(.-)" .. etx .. "()"
     if (gReceiveBuffer:len() > 0) then
	  message, pos = string.match(gReceiveBuffer, pattern) 
	  if (message == nil) then
	   --[[
		  pattern = "(PJ_OK)()"
		  message, pos = string.match(gReceiveBuffer, pattern) 
		  if (message == nil) then
			 pattern = "(PJ_NG)()"
			 message, pos = string.match(gReceiveBuffer, pattern) 
			 if (message == nil) then
				pattern = "(PJACK)()"
				message, pos = string.match(gReceiveBuffer, pattern) 
				if (message == nil) then
				    pattern = "(PJNAK)()"
				    message, pos = string.match(gReceiveBuffer, pattern) 
				    if (message == nil) then
					   --LogInfo("Do not have a complete message")
					   return ""
				    end
				end
			 end
		  end
	   --]]
	   
		  pattern = "(.....)()"	--PJ_OK, PJ_NG, PJACK, PJNAK
		  message, pos = string.match(gReceiveBuffer, pattern) 
		  if (message == nil) then
			 --LogInfo("Do not have a complete message")
			 return ""
		  end
		  
	  end
	  gReceiveBuffer = gReceiveBuffer:sub(pos)
     end

	return message
	
end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)
	if (type(message) == "table") then
		LogTrace("HandleMessage. Message is ==>")
		LogTrace(message)
		LogTrace("<==")
	else
		LogTrace("HandleMessage. Message is ==>%s<==", message)
	end

	if (message == "PJ_OK") then 
	   DEV_MSG.PJ_OK()
	   return
	end
	
	if (message == "PJNAK") then 
	   DEV_MSG.PJNAK()
	   return
	end	
	
     name, value = string.match(message, "(..)(.)") -- name parsed from Message

	name = name or message
	value = tonumber(value)
	value = value or ""	

	DispatchMessage(name, value)
	
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LogInfo("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhanded command = " .. MsgKey)
	end
end

function DEV_MSG.PJ_OK()
	LogTrace("PJ_OK()")
	QueuePriority1Command("PJREQ", "PJREQ")
end

function DEV_MSG.PJNAK()
	LogTrace("PJNAK()")
	QueuePriority1Command("RebootNetwork", "LSRS")
end

--[[
TODO: Create DEV_MSG functions for all messages to call Notifies.
	  Sample functions are included below for all applicable notifications.
--]]

function DEV_MSG.IP(value)
	LogTrace("DEV_MSG.INPUT(), value = " .. value)
	local input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
     local bindingID = tInputConnMapByName[tInputResponseMap[value]].BindingID 		
	LogInfo("INPUT_CHANGED, input = " .. tInputResponseMap[value])
	
	--all connections
	gProjectorProxy:dev_InputChanged(bindingID, input + 3000)		--audio
	
     --video only connections
	gProjectorProxy:dev_InputChanged(bindingID, input + 1000)		--video
	
end

function DEV_MSG.PW(value)
     gPowerEventTimeElapsed = os.time() - gPowerEventStarted 
	LogTrace("DEV_MSG.POWER(), value = " .. value .. "[Time Elpased: " .. gPowerEventTimeElapsed .. " seconds since " .. gLastPowerEvent .. "]")

	--[[
	TODO: "00" & "01" & "02" & "03" & "04" values will need to be edited based upon  
		the device protocol values, indicating if the device is "off", "on", 
		in a transitional state (i.e. Warming up, Cooling down) or in an error State.
		Then modify the if/then code block below to match the desired states.
	--]]
	if (value == 0) then
		if (gProjector_Operating_Status ~= "OFF") then
			gProjector_Operating_Status = "OFF"
			C4:SetVariable('Projector_Operating_Status',"OFF")
			gProjectorProxy:dev_PowerOff()
	     end
		
		if (gPowerOnDelayTimer:TimerStarted()) then
		    --a power on was intiated after this power off
		    local command_delay =  2	--tonumber(Properties["Power Off Delay Seconds"])
		    local delay_units = "SECONDS"
		    --OnPowerOnDelayTimerExpired()
		    gPowerOnDelayTimer:StartTimer(command_delay, delay_units)
		    if (gSetInputDelayTimer:TimerStarted()) then
			 gSetInputDelayTimer:StartTimer(120, delay_units)
		    end		    
		end	
		    
		
	elseif (value== 1) then
		if (gProjector_Operating_Status ~= "ON") then
			gProjector_Operating_Status = "ON"
			C4:SetVariable('Projector_Operating_Status',"ON")
			gProjectorProxy:dev_PowerOn()
	     end
		
		local command_delay =  2	--tonumber(Properties["Power On Delay Seconds"])
		local delay_units = "SECONDS"
		if (gPowerOffDelayTimer:TimerStarted()) then
		    --a power off was intiated after this power on
		    --OnPowerOffDelayTimerExpired()
		    gPowerOffDelayTimer:StartTimer(command_delay, delay_units)
		else
		    if (gSetInputDelayTimer:TimerStarted()) then
			 gSetInputDelayTimer:StartTimer(command_delay, delay_units)
		    end
		    
		end
		    
	elseif (value== 2) then	
		if (gProjector_Operating_Status ~= "COOLING_DOWN") then
			gProjector_Operating_Status = "COOLING_DOWN"
			C4:SetVariable('Projector_Operating_Status',"COOLING_DOWN")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
     elseif (value== 3) then	
		if (gProjector_Operating_Status ~= "WARMING_UP") then
			gProjector_Operating_Status = "WARMING_UP"
			C4:SetVariable('Projector_Operating_Status',"WARMING_UP")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== 4) then	
		if (gProjector_Operating_Status ~= "ERROR_CONDITION") then
			gProjector_Operating_Status = "ERROR_CONDITION"
			C4:SetVariable('Projector_Operating_Status',"ERROR_CONDITION")
		end
		--TODO: Handle the error...
		-- most likely want to send an Off Notification to the proxy
		-- so the power on will work after the error condition is satisfied
		gProjectorProxy:dev_PowerOff()		
	else
		LogWarn("DEV_MSG.POWER(): value not valid - " .. value)
	end		
end

function DEV_MSG.LAMP(value)
	LogTrace("Projector reports " .. value .. " Hours on the bulb")	
	C4:UpdateProperty('Lamp Hours', tostring(value))
	if (gLamp_Hours ~= value) then
		gLamp_Hours = value
		C4:SetVariable('Lamp_Hours',tostring(value))
	end	
end
	
function DEV_MSG.VOLUME(value)
	LogTrace("DEV_MSG.VOLUME(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = "" --mod 1000 value of Output Connection ID
	
	-- TODO: edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local deviceLevel = value
	local minDeviceLevel = -80
	local maxDeviceLevel = 20
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)		
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.MUTE(value)
	LogTrace("DEV_MSG.MUTE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == "01") then 
		state = "True"
	elseif (value == "00") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	gProjectorProxy:dev_MuteChanged(output, state)
end		

function DEV_MSG.BASS(value)
	LogTrace("DEV_MSG.BASS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BassLevelChanged(output, level)
end	

function DEV_MSG.TREBLE(value)
	LogTrace("DEV_MSG.TREBLE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Treble level is represented as a percentage value
	local level   
	
	gProjectorProxy:dev_TrebleLevelChanged(output, level)
end	

function DEV_MSG.BALANCE(value)
	LogTrace("DEV_MSG.BALANCE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BalanceLevelChanged(output, level)
end	

function DEV_MSG.LOUDNESS(value)
	LogTrace("DEV_MSG.LOUDNESS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "state", Loudness state is represented as "True" or "False" (literal string, not boolean)
	local state  
	gProjectorProxy:dev_LoudnessChanged(output, state)
end
	
 end)
package.preload['device_specific_commands'] = (function (...)
--[[=============================================================================
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_specific_commands = "2015.03.31"
end

--[[=============================================================================
    ExecuteCommand Code

    Define any functions for device specific commands (EX_CMD.<command>)
    received from ExecuteCommand that need to be handled by the driver.
===============================================================================]]
--function EX_CMD.NEW_COMMAND(tParams)
--	LogTrace("EX_CMD.NEW_COMMAND")
--	LogTrace(tParams)
--end

----------Start New 3.5 Commands added 2021------------
function EX_CMD.ContentType(tParams)
     tContentType = {
	   ["Auto"]	=	"0",
	   ["SDR"]	=	"1",
	   ["HDR10"]	=	"3",
	   ["HLG"]	=	"4",
     }
     local mode = tParams["Mode"]
     LogTrace("ContentType, Mode = " .. mode)
     local command = "PMCT" .. tContentType[mode]
	PackAndQueueCommand("ContentType, Mode = " .. mode, command)
end

function EX_CMD.ContentTypeAutoTransition(tParams)
     tContentTypeAutoTransition = {
	   ["SDR"]	=	"1",
	   ["HDR10"]	=	"3",
	   ["HLG"]	=	"4",
     }
     local mode = tParams["Mode"]
     LogTrace("ContentTypeAutoTransition, Mode = " .. mode)
     local command = "PMAT" .. tContentTypeAutoTransition[mode]
	PackAndQueueCommand("ContentTypeAutoTransition, Mode = " .. mode, command)
end

function EX_CMD.TheaterOptimizer(tParams)
     tTheaterOptimizer = {
	   ["Off"]	=	"0",
	   ["On"]		=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("TheaterOptimizer, Mode = " .. mode)
     local command = "PMNM" .. tTheaterOptimizer[mode]
	PackAndQueueCommand("TheaterOptimizer, Mode = " .. mode, command)
end

function EX_CMD.TheaterOptimizerLevel(tParams)
     tTheaterOptimizerLevel = {
	   ["Low"]	=	"1",
	   ["Mid"]	=	"2",
	   ["High"]	=	"3",
     }
     local mode = tParams["Mode"]
     LogTrace("TheaterOptimizerLevel, Mode = " .. mode)
     local command = "PMNL" .. tTheaterOptimizerLevel[mode]
	PackAndQueueCommand("TheaterOptimizerLevel, Mode = " .. mode, command)
end

function EX_CMD.TheaterOptimizerProcessing(tParams)
     tTheaterOptimizerProcessing = {
	   ["-"]				=	"0",
	   ["Processing Start"]	=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("TheaterOptimizerProcessing, Mode = " .. mode)
     local command = "PMNL" .. tTheaterOptimizerProcessing[mode]
	PackAndQueueCommand("TheaterOptimizerProcessing, Mode = " .. mode, command)
end

function EX_CMD.ColorProfileAutoTransition(tParams)
     tColorProfileAutoTransition = {
	   ["Off"]			=	"00",
	   ["Film1"]			=	"01",
	   ["Film2"]			=	"02",
	   ["BT.709"]			=	"03",
	   ["Cinema"]			=	"04",
	   ["Anime"]			=	"06",
	   ["Video"]			=	"08",
	   ["HDR"]			=	"0A",
	   ["BT.2020-Wide"]		=	"0B",
	   ["THX"]			=	"0D",
	   ["Custom1"]			=	"0E",
	   ["Custom2"]			=	"0F",
	   ["Custom3"]			=	"10",
	   ["Custom4"]			=	"11",
	   ["Custom5"]			=	"12",	--Pana_PQ_HL
	   ["Pana_PQ_HL"]		=	"12",
	   ["DCI"]			=	"21",	
	   ["Custom6"]			=	"22",	--Pana_PQ_BL
	   ["Pana_PQ_BL"]		=	"22",
	   ["BT.2020-Normal"]	=	"24",
	   ["Off-Wide"]		=	"25",
     }
     local mode = tParams["Mode"]
     LogTrace("ColorProfileAutoTransition, Mode = " .. mode)
     local command = "PMAF" .. tColorProfileAutoTransition[mode]
	PackAndQueueCommand("ColorProfileAutoTransition, Mode = " .. mode, command)
end

function EX_CMD.SDR2DPictureMode(tParams)
	 tSDR2DPictureMode = {
	   ['Natural'] 	= "1", 
	   ["User1"]		= "2",
	   ["User2"]		= "3",
	   ["User3"]		= "4",
	   ['Cinema'] 		= "8", 
	   ['Film'] 		= "9", 
	   ['THX'] 		= "E", 
	   ['Last Setting'] = "F", 
     }
     local mode = tParams["Mode"]
     LogTrace("SDR2DPictureMode, Mode = " .. mode)
     local command = "ISS2" .. tSDR2DPictureMode[mode]
	PackAndQueueCommand("SDR2DPictureMode, Mode = " .. mode, command)
end

function EX_CMD.SDR3DPictureMode(tParams)
	 tSDR3DPictureMode = {
	   ['Natural'] 	= "1", 
	   ["User1"]		= "2",
	   ["User2"]		= "3",
	   ["User3"]		= "4",
	   ['Cinema'] 		= "8", 
	   ['Film'] 		= "9", 
	   ['THX'] 		= "E", 
	   ['Last Setting'] = "F", 
     }
     local mode = tParams["Mode"]
     LogTrace("SDR3DPictureMode, Mode = " .. mode)
     local command = "ISS3" .. tSDR3DPictureMode[mode]
	PackAndQueueCommand("SDR3DPictureMode, Mode = " .. mode, command)
end

function EX_CMD.ScreenSize(tParams)
     local size = tParams["Size"]
     LogTrace("ScreenSize, Size = " .. size)
     local command = "INNS" .. size
	PackAndQueueCommand("ScreenSize, Size = " .. size, command)
end

function EX_CMD.ScreenGain(tParams)
     local gain = tParams["Gain"]
     LogTrace("ScreenGain, Gain = " .. gain)
     local command = "INNG" .. gain
	PackAndQueueCommand("ScreenGain, Gain = " .. gain, command)
end

-----------End New 3.5 Commands added 2021-------------


-----------Start New 2018 Commands------------

function EX_CMD.IntelligentLensAperture(tParams)
     tIntelligentLensAperture = {
	   ["Off"] 		= 	"00", 
	   ["Auto1"]		=	"01", 
	   ["Auto2"]		=	"02",
     }
     local mode = tParams["Mode"]
     LogTrace("IntelligentLensAperture, Mode = " .. mode)
     local command = "PMDI" .. tIntelligentLensAperture[mode]
	PackAndQueueCommand("IntelligentLensAperture, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.ESHIFT8K(tParams)
     tESHIFT = {
	   ["Off"]	=	"0",
	   ["On"]		=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("ESHIFT8K, Mode = " .. mode)
     local command = "PMUS" .. tESHIFT[mode]
	PackAndQueueCommand("ESHIFT8K, Mode = " .. mode, command)
end

function EX_CMD.AutoToneMapping(tParams)
     tAutoToneMapping = {
	   ["Off"]	=	"0",
	   ["On"]		=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("AutoToneMapping, Mode = " .. mode)
     local command = "PMTM" .. tAutoToneMapping[mode]
	PackAndQueueCommand("AutoToneMapping, Mode = " .. mode, command)
end

function EX_CMD.HDR10PictureMode(tParams)
     tHDR10PictureMode = {
	   ["Off"]			=	"0",
	   ["HDR10"]			=	"1",
	   ["User1"]			=	"2",
	   ["User2"]			=	"3",
	   ["User3"]			=	"4",
	   ["User4"]			=	"5",
	   ["User5"]			=	"6",
	   ["User6"]			=	"7",
	   ["Frame Adapt HDR"]	=	"8",
	   ["Pana PQ"]			=	"9",
	   ["Last Setting"]		=	"F",
     }
     local mode = tParams["Mode"]
     LogTrace("HDR10PictureMode, Mode = " .. mode)
     local command = "ISHD" .. tHDR10PictureMode[mode]
	PackAndQueueCommand("HDR10PictureMode, Mode = " .. mode, command)
end

function EX_CMD.HLGPictureMode(tParams)
     tHLGPictureMode = {
	   ["Off"]		=	"0",
	   ["HLG"]		=	"1",
	   ["USER1"]		=	"2",
	   ["USER2"]		=	"3",
	   ["USER3"]		=	"4",
	   ["USER4"]		=	"5",
	   ["USER5"]		=	"6",
	   ["USER6"]		=	"7",
	   ["Last Setting"]	=	"F",
	   
     }
     local mode = tParams["Mode"]
     LogTrace("HLGPictureMode, Mode = " .. mode)
     local command = "ISHL" .. tHLGPictureMode[mode]
	PackAndQueueCommand("HLGPictureMode, Mode = " .. mode, command)
end

function EX_CMD.OriginalResolution(tParams)
     tMode = {
	   ["Off"]	=	"0",
	   ["On"]		=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("OriginalResolution, Mode = " .. mode)
     local command = "PMRP" .. tMode[mode]
	PackAndQueueCommand("OriginalResolution, Mode = " .. mode, command)
end
-----------End New 2018 Commands------------

function EX_CMD.PictureMode(tParams)
     tPictureMode = {
	   ["Film"] 			= 	"00", 
	   ["Cinema"]			=	"01", 
	   ["Animation"]		=	"02",
	   ["Natural"]			=	"03", 
	   ["HDR10"]			=	"04", 
	   ["Dynamic"]			=	"05",
	   ["THX"] 			=	"06", 
	   ["isf-Day"]			=	"07", 
	   ["isf-Night"]		=	"08", 
	   ["THX-Bright"] 		=	"09", 
	   ["THX-Dark"]		=	"0A", 
	   ["Frame Adapt HDR"]	=	"0B", 
	   ["User1"]  			=	"0C", 
	   ["User2"]  			=	"0D", 
	   ["User3"]  			=	"0E", 
	   ["User4"]  			=	"0F", 
	   ["User5"]  			=	"10",  
	   ["User6"]  			=	"11",
	   ["HLG"]			=	"14", 
	   ["Pana PQ"]			=	"16", 
     }
     local mode = tParams["Mode"]
     LogTrace("PictureMode, Mode = " .. mode)
     local command = "PMPM" .. tPictureMode[mode]
	PackAndQueueCommand("PictureMode, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.ColorManagement(tParams)
     tColorManagement = {
	   ["Off"]	=	"0",
	   ["On"]	=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("ColorManagement, Mode = " .. mode)
     local command = "PMCB" .. tColorManagement[mode]
	PackAndQueueCommand("ColorManagement, Mode = " .. mode, command)
end

function EX_CMD.ColorProfile(tParams)
     tColorProfile = {
	   ["Off"]			=	"00",
	   ["Film1"]			=	"01",
	   ["Film2"]			=	"02",
	   ["BT.709"]			=	"03",
	   ["Cinema"]			=	"04",
	   ["Cinema2"]			=	"05",
	   ["Anime"]			=	"06",
	   ["Anime2"]			=	"07",
	   ["Video"]			=	"08",
	   ["x.v.Color"]		=	"09",
	   ["HDR"]			=	"0A",
	   ["BT.2020"]			=	"0B",
	   ["BT.2020-Wide"]			=	"0B",
	   ["3D Cinema"]		=	"0C",
	   ["THX"]			=	"0D",
	   ["Custom1"]			=	"0E",
	   ["Custom2"]			=	"0F",
	   ["Custom3"]			=	"10",
	   ["Custom4"]			=	"11",
	   ["Custom5"]			=	"12",
	   ["Pana_PQ_HL"]		=	"12",
	   ["Film3"]			=	"13",
	   ["3D Video"]		=	"14",
	   ["3D Animation"]		=	"15",
	   ["Frame Adapt HDR"]	=	"17",
	   ["3D Film"]			=	"1E",
	   ["3D THX"]			=	"20",
	   ["DCI"]			=	"21",
	   ["Custom6"]			=	"22",
	   ["Pana_PQ_BL"]		=	"22",
	   ["BT.2020-Normal"]	=	"24",
	   ["Off-Wide"]		=	"25",
	   ["Auto"]			=	"26"
     }
     local mode = tParams["Mode"]
     LogTrace("ColorProfile, Mode = " .. mode)
     local command = "PMPR" .. tColorProfile[mode]
	PackAndQueueCommand("ColorProfile, Mode = " .. mode, command, 5, "SECONDS")
	
	--[[
	NOTE: Only the parameter that follows Picture Mode is effective. 
	[Example] 	Picture Mode = Natural Video(0x38) -> Accepted, Anime1(0x36) -> Rejected
				Picture Mode = Film Film1(0x31) -> Accepted, Standard(0x33) -> Rejected
	If the corresponded parameter is only one, PJ ignores setting command.
	--]]
end

function EX_CMD.ColorTemp(tParams)
     tColorTemp = {
	   ["5500K"]		=	"00",
	   ["6500K"]		=	"02",
	   ["7500K"]		=	"04",
	   ["9300K"]		=	"08",
	   ["High Bright"]	=	"09",
	   ["Custom1"]		=	"0A",
	   ["Custom2"]		=	"0B",
	   ["HDR"]		=	"0C",
	   ["HDR10"]		=	"0C",
	   ["Xenon1"]		=	"0D",
	   ["Xenon2"]		=	"0E",
	   ["HLG"]		=	"14",
     }
     local mode = tParams["Mode"]
     LogTrace("ColorTemp, Mode = " .. mode)
     local command = "PMCL" .. tColorTemp[mode]
	PackAndQueueCommand("ColorTemp, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.ColorTempCorrection(tParams)
     tColorTempCorrection = {
	   ["5500K"]		=	"00",
	   ["6500K"]		=	"02",
	   ["7500K"]		=	"04",
	   ["9300K"]		=	"08",
	   ["High Bright"]	=	"09",
	   ["Xenon1"]		=	"0D",
	   ["Xenon2"]		=	"0E",
     }
     local mode = tParams["Mode"]
     LogTrace("ColorTempCorrection, Mode = " .. mode)
     local command = "PMCC" .. tColorTempCorrection[mode]
	PackAndQueueCommand("ColorTempCorrection, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.GammaTable(tParams)
     tGammaTable = {
	   ["2.2"]			=	"00",
	   ["Cinema1"]			=	"01",
	   ["Cinema2"]			=	"02",
	   ["Custom1"]			=	"04",
	   ["Custom2"]			=	"05",
	   ["Custom3"]			=	"06",
	   ["Hybrid Log"]		=	"07",
	   ["HDR(HLG)"]		=	"07",
	   ["2.4"]			=	"08",
	   ["2.6"]			=	"09",
	   ["Film1"]			=	"0A",
	   ["Film2"]			=	"0B",
	   ["ST.2084"]			=	"0C",
	   ["HDR(PQ)"]			=	"0C",
	   ["Frame Adapt HDR"]	=	"0D",
	   ["Pana PQ"]			=	"0D",
	   ["THX"]			=	"10",
	   ["HDR-Auto"]		=	"15",
	   
     }
     local mode = tParams["Mode"]
     LogTrace("GammaTable, Mode = " .. mode)
     local command = "PMGT" .. tGammaTable[mode]
	PackAndQueueCommand("GammaTable, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.GammaCorrection(tParams)
     tGammaCorrection = {
	   ["Normal"]		=	"00",
	   ["Cinema1"]		=	"01",
	   ["Cinema2"]		=	"02",
	   ["Import"]		=	"04",
	   ["Import1"]		=	"04",
	   ["1.8"]		=	"05",
	   ["1.9"]		=	"06",
	   ["2.0"]		=	"07",
	   ["2.1"]		=	"08",
	   ["2.2"]		=	"09",
	   ["2.3"]		=	"0A",
	   ["2.4"]		=	"0B",
	   ["2.5"]		=	"0C",
	   ["2.6"]		=	"0D",
	   ["Film1"]		=	"0E",
	   ["Film2"]		=	"0F",
	   ["Import2"]		=	"12",
	   ["Import3"]		=	"13",
	   ["Hybrid Log"]	=	"14",
	   ["HDR(HLG)"]	=	"14",
	   ["ST.2084"]		=	"15",
	   ["HDR(PQ)"]		=	"15",
     }
	
     local mode = tParams["Mode"]
     LogTrace("GammaCorrection, Mode = " .. mode)
     local command = "PMGC" .. tGammaCorrection[mode]
	PackAndQueueCommand("GammaCorrection, Mode = " .. mode, command, 5, "SECONDS")
end

function EX_CMD.DynamicControl(tParams)
     tDynamicControl = {
	   ["Off"]		=	"0",
	   ["Mode1"]	=	"1",
	   ["Mode2"]	=	"2",
     }
     local mode = tParams["Mode"]
     LogTrace("DynamicControl, Mode = " .. mode)
     local command = "PMDC" .. tDynamicControl[mode]
	PackAndQueueCommand("DynamicControl, Mode = " .. mode, command)
end

function EX_CMD.LowLatency(tParams)
     tLowLatency = {
	   ["Off"]	=	"0",
	   ["On"]	=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("LowLatency, Mode = " .. mode)
     local command = "PMLL" .. tLowLatency[mode]
	PackAndQueueCommand("LowLatency, Mode = " .. mode, command)
end

function EX_CMD.GraphicMode(tParams)
     tGraphicMode = {
	   ["2K"]			=	"0",
	   ["4K"]			=	"1",
	   ["Standard"]	=	"0",
	   ["High-res"]	=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("GraphicMode, Mode = " .. mode)
     local command = "PMGM" .. tGraphicMode[mode]
	PackAndQueueCommand("GraphicMode, Mode = " .. mode, command)
end

function EX_CMD.Anamorphic(tParams)
     tAnamorphic = {
	   ["Off"]	=	"0",
	   ["A"]		=	"1",
	   ["B"]		=	"2",
	   ["C"]		=	"3",
	   ["D"]		=	"4",
     }
     local mode = tParams["Mode"]
     LogTrace("Anamorphic, Mode = " .. mode)
     local command = "INVS" .. tAnamorphic[mode]
	PackAndQueueCommand("Anamorphic, Mode = " .. mode, command)
end

function EX_CMD.HDMI2D3D(tParams)
     tHDMI2D3D = {
	   ["2D"]				=	"0",
	   ["Auto"]				=	"1",
	   ["Side by Side"]		=	"3",
	   ["Top and Bottom"]	=	"4",
     }
     local mode = tParams["Mode"]
     LogTrace("HDMI2D3D, Mode = " .. mode)
     local command = "IS3D" .. tHDMI2D3D[mode]
	PackAndQueueCommand("HDMI2D3D, Mode = " .. mode, command)
end

function EX_CMD._4Keshift(tParams)
     t4Keshift = {
	   ["Off"]		=	"0",
	   ["On "]		=	"1",
     }
     local mode = tParams["Mode"]
     LogTrace("4Keshift, Mode = " .. mode)
     local command = "PMUS" .. t4Keshift[mode]
	PackAndQueueCommand("4Keshift, Mode = " .. mode, command)
end

function EX_CMD.Source(tParams)
     tSource = {
	   ["480p"]				=	"02",
	   ["576p"]				=	"03",
	   ["720p50"]			=	"04",
	   ["720p60"]			=	"05",
	   ["1080i50"]			=	"06",
	   ["1080i60"]			=	"07",
	   ["1080p24"]			=	"08",
	   ["1080p50"]			=	"09",
	   ["1080p60"]			=	"0A",
	   ["No Signal"]		=	"0B",
	   ["720p 3D"]			=	"0C",
	   ["1080i 3D"]			=	"0D",
	   ["1080p 3D"]			=	"0E",
	   ["4K(4096)60"]		=	"10",
	   ["4K(4096)50"]		=	"11",
	   ["4K(4096)30"]		=	"12",
	   ["4K(4096)25"]		=	"13",
	   ["4K(4096)24"]		=	"14",
	   ["4K(3840)60"]		=	"15",
	   ["4K(3840)50"]		=	"16",
	   ["4K(3840)30"]		=	"17",
	   ["4K(3840)25"]		=	"18",
	   ["4K(3840)24"]		=	"19",
	   
	   ["1080p25"]			=	"1C",
	   ["1080p30"]			=	"1D",
	   ["2048x1080 p24"]	=	"1E",
	   ["2048x1080 p25"]	=	"1F",
	   ["2048x1080 p30"]	=	"20",
	   ["2048x1080 p50"]	=	"21",
	   ["2048x1080 p60"]	=	"22",
	   ["3840x2160 p120"]	=	"23",
	   ["4096x2160 p120"]	=	"24",
	   
	   ["VGA(640x480)"]		=	"25",
	   ["SVGA(800x600)"]	=	"26",
	   ["XGA(1024x768)"]	=	"27",
	   ["SXGA(1280x1024)"]	=	"28",
	   ["WXGA(1280x768)"]	=	"29",
	   ["WXGA+(1440x900)"]	=	"2A",
	   ["WSXGA+(1680x1050)"]	=	"2B",
	   ["WUXGA(1920x1200)"]	=	"2C",
	   ["WXGA(1280x800)"]	=	"2D",
	   ["FWXGA(1366x768)"]	=	"2E",
	   ["WXGA++(1600x900)"]	=	"2F",
	   ["UXGA(1600x1200)"]	=	"30",
	   ["QXGA"]			=	"31",
	   ["WQXGA"]			=	"32",
	   
     }
     local mode = tParams["Mode"]
     LogTrace("Source, Mode = " .. mode)
     local command = "IFIS" .. tSource[mode]
	PackAndQueueCommand("Source, Mode = " .. mode, command, 3, "SECONDS")
end

function EX_CMD.Aspect(tParams)
     tAspect = {
	   ["4:3"]			=	"0",
	   ["16:9"]			=	"1",
	   ["Zoom"]			=	"2",
	   ["Auto"]			=	"3",
	   ["Just"]			=	"4",
	   ["Native"]			=	"4",
	   ["Full"]			=	"5",
     }
     local mode = tParams["Mode"]
     LogTrace("Aspect, Mode = " .. mode)
     local command = "ISAS" .. tAspect[mode]
	PackAndQueueCommand("Aspect, Mode = " .. mode, command)
end

function EX_CMD.LensMemory(tParams)
     tLensMemory = {
	   ["Memory 1"]		=	"0",
	   ["Memory 2"]		=	"1",
	   ["Memory 3"]		=	"2",
	   ["Memory 4"]		=	"3",
	   ["Memory 5"]		=	"4",
	   ["Memory 6"]		=	"5",
	   ["Memory 7"]		=	"6",
	   ["Memory 8"]		=	"7",
	   ["Memory 9"]		=	"8",
	   ["Memory 10"]	=	"9",
     }
     local mode = tParams["Mode"]
     LogTrace("LensMemory, Mode = " .. mode)
     local command = "INML" .. tLensMemory[mode]
	PackAndQueueCommand("LensMemory, Mode = " .. mode, command)
end

function EX_CMD.InstallationMode(tParams)
     tInstallationMode = {
	   ["Mode 1"]		=	"0",
	   ["Mode 2"]		=	"1",
	   ["Mode 3"]		=	"2",
	   ["Mode 4"]		=	"3",
	   ["Mode 5"]		=	"4",
	   ["Mode 6"]		=	"5",
	   ["Mode 7"]		=	"6",
	   ["Mode 8"]		=	"7",
	   ["Mode 9"]		=	"8",
	   ["Mode 10"]		=	"9",
     }
     local mode = tParams["Mode"]
     LogTrace("InstallationMode, Mode = " .. mode)
     local command = "INML" .. tInstallationMode[mode]
	PackAndQueueCommand("InstallationMode, Mode = " .. mode, command)
end

function EX_CMD.ClearMotionDrive(tParams)
	 tClearMotionDrive = {
		['Off'] 				= "0", 
		['Low'] 				= "3", 
		['High'] 				= "4", 
		['Inverse Telecine'] 	= "5", 
     }
     local mode = tParams["Mode"]
     LogTrace("ClearMotionDrive, Mode = " .. mode)
	 local command = "PMCM" .. tClearMotionDrive[mode]
	PackAndQueueCommand("ClearMotionDrive, Mode = " .. mode, command)
end

function EX_CMD.MotionEnhance(tParams)
	 tMotionEnhance = {
		['Off'] 				= "0", 
		['Low'] 				= "1", 
		['High'] 				= "2", 
     }
     local mode = tParams["Mode"]
     LogTrace("MotionEnhance, Mode = " .. mode)
	 local command = "PMME" .. tMotionEnhance[mode]
	PackAndQueueCommand("MotionEnhance, Mode = " .. mode, command)
end

function EX_CMD.LDPower(tParams)
	 tLDPower = {
		['Normal'] 				= "0", 
		['High'] 				= "1", 
		['Mid'] 				= "2", 
     }
     local mode = tParams["Mode"]
     LogTrace("LDPower, Mode = " .. mode)
	 local command = "PMLP" .. tLDPower[mode]
	PackAndQueueCommand("LDPower, Mode = " .. mode, command)
end

function EX_CMD.Mask(tParams)
	 tMask = {
		['Off'] 	= "0", 
		['On'] 		= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("Mask, Mode = " .. mode)
	 local command = "ISMA" .. tMask[mode]
	PackAndQueueCommand("Mask, Mode = " .. mode, command)
end


function EX_CMD.HighAltitude(tParams)
	 tHighAltitude = {
		['Off'] 	= "0", 
		['On'] 		= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("HighAltitude, Mode = " .. mode)
	 local command = "INHA" .. tHighAltitude[mode]
	PackAndQueueCommand("HighAltitude, Mode = " .. mode, command)
end

function EX_CMD.BackColor(tParams)
	 tBackColor = {
		['Blue'] 		= "0", 
		['Black'] 	= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("BackColor, Mode = " .. mode)
     local command = "DSBC" .. tBackColor[mode]
	PackAndQueueCommand("BackColor, Mode = " .. mode, command)
end

function EX_CMD.MenuPosition(tParams)
	 tMenuPosition = {
		['Left-Top'] 		= "0", 
		['Right-Top'] 		= "1", 
		['Center'] 			= "2", 
		['Left-Bottom'] 	= "3", 
		['Right-Bottom'] 	= "4", 
		['Left'] 			= "5",
		['Right'] 			= "6",  
     }
     local mode = tParams["Mode"]
     LogTrace("MenuPosition, Mode = " .. mode)
	 local command = "DSMP" .. tMenuPosition[mode]
	PackAndQueueCommand("MenuPosition, Mode = " .. mode, command)
end

function EX_CMD.SourceDisplay(tParams)
	 tSourceDisplay = {
		['Off'] 	= "0", 
		['On'] 		= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("SourceDisplay, Mode = " .. mode)
	 local command = "DSSD" .. tSourceDisplay[mode]
	PackAndQueueCommand("SourceDisplay, Mode = " .. mode, command)
end

function EX_CMD.OffTimer(tParams)
	 tOffTimer = {
		['Off'] 		= "0", 
		['1 Hour'] 		= "1", 
		['2 Hours'] 	= "2", 
		['3 Hours'] 	= "3", 
		['4 Hours'] 	= "4", 
     }
     local mode = tParams["Mode"]
     LogTrace("OffTimer, Mode = " .. mode)
	 local command = "FUOT" .. tOffTimer[mode]
	PackAndQueueCommand("OffTimer, Mode = " .. mode, command)
end

function EX_CMD.Trigger(tParams)
	 tTrigger = {
		['Off'] 		= "0", 
		['Power'] 		= "1", 
		['Anamorphic'] 	= "2", 
		['Inst. 1'] 	= "3", 
		['Inst. 2'] 	= "4", 
		['Inst. 3'] 	= "5", 
		['Inst. 4'] 	= "6", 
		['Inst. 5'] 	= "7", 
		['Inst. 6'] 	= "8", 
		['Inst. 7'] 	= "9", 
		['Inst. 8'] 	= "A", 
		['Inst. 9'] 	= "B", 
		['Inst. 10'] 	= "C", 
     }
     local mode = tParams["Mode"]
     LogTrace("Trigger, Mode = " .. mode)
     local command = "FUTR" .. tTrigger[mode]
	PackAndQueueCommand("Trigger, Mode = " .. mode, command)
end

function EX_CMD.EcoMode(tParams)
	 tEcoMode = {
		['Off'] 	= "0", 
		['On'] 		= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("EcoMode, Mode = " .. mode)
	 local command = "FUEM" .. tEcoMode[mode]
	PackAndQueueCommand("EcoMode, Mode = " .. mode, command)
end


function EX_CMD.DeepColor(tParams)
	 tDeepColor = {
		['8 bit'] 		= "0", 
		['10 bit'] 		= "1", 
		['12 bit'] 		= "2", 
     }
     local mode = tParams["Mode"]
     LogTrace("DeepColor, Mode = " .. mode)
	 local command = "IFDC" .. tDeepColor[mode]
	PackAndQueueCommand("DeepColor, Mode = " .. mode, command)
end

function EX_CMD.ColorSpace(tParams)
	 tColorSpace = {
		['RGB'] 						= "0", 
		['YUV'] 							= "1", 
     }
     local mode = tParams["Mode"]
     LogTrace("ColorSpace, Mode = " .. mode)
	 local command = "IFXV" .. tColorSpace[mode]
	PackAndQueueCommand("ColorSpace, Mode = " .. mode, command)
end

function EX_CMD.Colorimetry(tParams)
	 tColorimetry = {
		['No Data'] 						= "0", 
		['BT.601'] 							= "1", 
		['BT.709'] 							= "2", 
		['xvYCC601'] 						= "3", 
		['xvYCC709'] 						= "4", 
		['sYCC601'] 						= "5", 
		['Adobe YCC601'] 					= "6", 
		['Adobe RGB'] 						= "7", 
		['BT.2020 Constant Luminance'] 		= "8", 
		['BT.2020 Non-Constant Luminance'] 	= "9", 
		['sRGB'] 							= "A", 
     }
     local mode = tParams["Mode"]
     LogTrace("Colorimetry, Mode = " .. mode)
	 local command = "IFCM" .. tColorimetry[mode]
	PackAndQueueCommand("Colorimetry, Mode = " .. mode, command)
end

function EX_CMD.HDR(tParams)
	 tHDR = {
		['SDR'] 			= "0", 
		['HDR'] 			= "1", 
		['SMPTE ST 2084'] 	= "2", 
		['Hybrid Log'] 	= "3",
		['None'] 			= "F", 
     }
     local mode = tParams["Mode"]
     LogTrace("HDR, Mode = " .. mode)
     local command = "IFHR" .. tHDR[mode]
	PackAndQueueCommand("HDR, Mode = " .. mode, command)
end

function EX_CMD.HDRLevel(tParams)
	 tLevel = {
	     ['Auto'] 		= "0",
		['-2'] 		= "1", 
		['-1'] 		= "2", 
		['0'] 		= "3", 
		['1'] 		= "4", 
		['2'] 		= "5", 
     }
     local level = tParams["Level"]
     LogTrace("HDR Level, Level = " .. level)
     local command = "PMHL" .. tLevel[level]
	PackAndQueueCommand("HDR Level, Level = " .. level, command)
end

function EX_CMD.HDRProcessing(tParams)
	 tHDRProcessing = {
		['Static'] 		= "1", 
		['Frame by Frame'] 	= "2", 
		['Scene by Scene'] 	= "3",  
     }
     local mode = tParams["Mode"]
     LogTrace("HDRProcessing, Mode = " .. mode)
     local command = "PMHP" .. tHDRProcessing[mode]
	PackAndQueueCommand("HDRProcessing, Mode = " .. mode, command)
end
 end)
package.preload['actions'] = (function (...)
--[[=============================================================================
    Lua Action Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.actions = "2015.03.02"
end

-- TODO: Create a function for each action defined in the driver

function LUA_ACTION.TemplateVersion()
	TemplateVersion()
end
 end)
package.preload['common.c4_utils'] = (function (...)
--[[=============================================================================
    Helper functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_utils = "2015.03.02"
end

--[[=============================================================================
    AsciiToBCD(InString)

    Description
    Convert an ascii string to a binary coded decimal. Each decimal digit is
    stored in one byte, with the lower four bits encoding the digit in BCD form.

    Parameters
    InString(string) - Ascii string that is to be converted into bcd

    Returns
    The binary coded decimal
===============================================================================]]
function AsciiToBCD(InString)
	local WorkVal = 0
	local RetValStr = ""
	local DoingHighNybble = false
	local WorkStr = ((#InString % 2) == 0) and (InString) or ("0" .. InString)	-- make sure length is an even number

	for CharCount = 1, #WorkStr do
		local NumVal = tonumber(WorkStr:sub(CharCount, CharCount))

		WorkVal = bit.lshift(WorkVal, 4) + NumVal
		if (DoingHighNybble) then
			RetValStr = RetValStr .. string.char(WorkVal)
			WorkVal = 0
		end

		DoingHighNybble = (not DoingHighNybble)
	end

	return RetValStr
end

--[[=============================================================================
    BCDToAscii(InByte)

    Description
    Convert an BCD string to an ascii string.

    Parameters
    InByte(string) - Binary coded decimal that is to be converted into ascii

    Returns
    The ascii string
===============================================================================]]
function BCDToAscii(InByte)
	return tostring(bit.rshift(InByte, 4)) .. tostring(bit.band(InByte, 0x0F))
end

--[[=============================================================================
    MakeXMLNode(Tag, Value)

    Description
    Create an Xml element

    Parameters
    Tag(string)   - The Xml elements name
    Value(string) - The Xml elements value

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLNode(Tag, Value)
	return "<" .. Tag .. ">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)

    Description
    Create an Xml element with an attribute

    Parameters
    Tag(string)       - The Xml elements name
    Value(string)     - The Xml elements value
    Attribute(string) - The attribute to be added to the Xml element
    AttrValue(string) - The value of the attribute to be added

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)
    return "<" .. Tag .. " " .. Attribute .. "=\"" .. AttrValue .. "\">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    StringFromUnicode(UnicodeString)

    Description
    Convert a unicode string

    Parameters
    UnicodeString(string) - The unicode string to be converted to ascii

    Returns
    The ascii representation of the unicode string
===============================================================================]]
function StringFromUnicode(UnicodeString)
	local RetVal = ""

	-- extract every other byte from the unicode string
	for Index = 2, #UnicodeString, 2 do
		RetVal = RetVal .. string.sub(UnicodeString, Index, Index)
	end

	return RetVal
end

--[[=============================================================================
    StringSplit(s)

    Description
    Splits a string into multiple strings at every space character

    Parameters
    s(string) - The string that is to be split into several strings at the
                space character

    Returns
    A table of strings containing all the seperate values in the given string
===============================================================================]]
function StringSplit(s)
	local delim = " "
	local result = {}

	for match in (s..delim):gmatch("(.-)"..delim) do
		table.insert(result, match)
	end

	return result
end

--[[=============================================================================
    toboolean(s)

    Description
    Returns a boolean representation of the given string

    Parameters
    s(string) - The string to evaluate the boolean representation from

    Returns
    The value true or false based on the given string
===============================================================================]]
function toboolean(s)
	local ret = nil
	if (s ~= nil) then
		if (type(s) == "boolean") then
			ret = s
		elseif (type(s) == "string") then
			ret = (string.lower(s) == "true")
		end
	end
	
	return ret
end

--[[=============================================================================
    Go(to, err, ...)

    Description
    Call a function with the given arguments if it exists or report the error

    Parameters
    to(string)  - The string to evaluate the boolean representation from
    err(string) - The error to report if the function does not exist
    ...         - Additional optional parameters for the function specified by
                  the "to" parameter

    Returns
    Nothing
===============================================================================]]
function Go(to, err, ...)
	if (type(to) == "function") then
		return to(...)
	else
		LogTrace(err)
	end
end

--[[=============================================================================
    IsEmpty(str)

    Description
    Identifies if the string given is nil or empty

    Parameters
    str(string) - The string to evaluate for the empty condition

    Returns
    True if the given value is empty, false otherwise
===============================================================================]]
function IsEmpty(str)
	return str == nil or str == ""
end

--[[=============================================================================
    ReverseTable(a)

    Description
    Reverse table entries (key=value, value=key)

    Parameters
    a(table) - The table to reverse

    Returns
    new reversed table
===============================================================================]]
function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function tonumber_loc(str, base)
  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

--[[=============================================================================
    HexToString(InString)

    Description
    Converts a string of Hex characters to a readable string of ASCII characters

    Parameters
    InString(string) - The string to be converted

    Returns
    A string showing the hex bytes of the InString
===============================================================================]]
function HexToString(InString)
	local RetVal = ""

	for Index = 1, #InString do
		RetVal = RetVal .. string.format("%02X ", InString:byte(Index))
	end
	return RetVal
end


--[[=============================================================================
    StringToHex(InString)

    Description
    Converts a string of ASCII characters to as string with the actual Hex bytes in them.
	Basically an array of hex bytes.

    Parameters
    InString(string) - The string to be converted

    Returns
    A string of hex bytes (really an array of hex values) 
===============================================================================]]
function StringToHex(InString)
	local RetVal = ""

	for HexByteString in string.gfind(InString, "%x%x") do
		RetVal = RetVal .. string.char(tonumber(HexByteString, 16))
	end
	return RetVal
end


 end)
package.preload['common.c4_ir_connection'] = (function (...)
--[[=============================================================================
    Base for an IR connection driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_ir_connection = "2015.03.02"
end

IRConnectionBase = inheritsFrom(DeviceConnectionBase)

function IRConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function IRConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("tSerConBase:Initialize")
	gControlMethod = "IR"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function IRConnectionBase:ControlMethod()
	return "IR"
end

function IRConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendIR(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("IR connection is not bound. Command not sent.")
	end
end
 end)
package.preload['common.c4_serial_connection'] = (function (...)
--[[=============================================================================
    Base for a serial connection driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_serial_connection = "2015.03.02"
end

SerialConnectionBase = inheritsFrom(DeviceConnectionBase)

function SerialConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function SerialConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "Serial"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function SerialConnectionBase:ControlMethod()
	return "Serial"
end

function SerialConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendToSerial(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("Not connected to serial. Command not sent.")
	end
end

function SerialConnectionBase:SendRaw(sData)
	C4:SendToSerial(self._BindingID, sData)
end


function SerialConnectionBase:ReceivedFromSerial(idBinding, sData)
	self:ReceivedFromCom(sData)
end
 end)
package.preload['lib.c4_queue'] = (function (...)
--[[=============================================================================
    c4_queue Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_queue = "2015.03.02"
end

c4_queue = inheritsFrom(nil)

function c4_queue:construct()
	-- entry table
	self._et = {first = 0, last = -1}
	self._maxSize = 0	-- no size limit
	self._name = ""

	local mt = getmetatable(self)
	if (mt ~= nil) then
		mt.__tostring = self.__tostring
	end
end

function c4_queue:__tostring()
	local tOutputString = {}
	table.insert(tOutputString, "--- Queue ---")
	if (not IsEmpty(self._name)) then
		table.insert(tOutputString, "  name = " .. tostring(self._name))
	end
	table.insert(tOutputString, "  first = " .. tostring(self._et.first))
	table.insert(tOutputString, "  last = " .. tostring(self._et.last))
	table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
	table.insert(tOutputString, "  maximum size = " .. self._maxSize)
	table.insert(tOutputString, "  next value = " .. tostring(self:value()))
	return table.concat(tOutputString, "\n")
end


-- push a value on the queue
function c4_queue:push(value, ...)
	local numItems = self._et.last - self._et.first + 1

	if ((self._maxSize <= 0) or (numItems < self._maxSize) ) then
		local last = self._et.last + 1
		self._et.last = last
		
		local interval = select(1, ...)
		local units = select(2, ...)
		local command_name = select(3, ...)
		self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}		
		--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
	else
		-- if addToQueue == true then push value to queue
		if (self:OnMaxSizeReached()) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}
			--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		end
	end
end

function c4_queue:OnMaxSizeReached()
	--LogTrace ("Max Size Reached - clear queue and push value to the queue (default).")
	local addToQueue = true

	self:clear()
	return (addToQueue)
end

-- pop a value from the queue
function c4_queue:pop()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace("Queue:pop(), queue is empty")
		return ""
	end

	local value = self._et[first]
	self._et[first] = nil        -- to allow garbage collection
	self._et.first = first + 1
	--LogTrace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)

	return value
end

-- clear queue
function c4_queue:clear()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace ("Queue:clear(), queue is empty")
		return ""
	end

	self._et = {first = 0, last = -1}
	--LogTrace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
	--LogTrace (self._et)
end

-- return value of first item
function c4_queue:value()
	local first = self._et.first

	if (first > self._et.last) then
		return ""
	else
		return self._et[first]
	end
end

-- return queue's maximum size
function c4_queue:MaxSize()
	return self._maxSize
end

-- return queue's maximum size
function c4_queue:SetMaxSize(size)
	self._maxSize = size
end

function c4_queue:SetName(name)
	self._name = name
end

-- return the queue's current size
function c4_queue:Size()
	return self._et.last - self._et.first + 1
end

-- is queue empty?
function c4_queue:empty()
	-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
	if (self._et.first > self._et.last) then
		return true
	else
		return false
	end
end

--[[
    c4_queue unit tests
--]]
function __test_c4_queue()
	require "test.C4Virtual"
	require "lib.c4_log"

	local LOG = c4_log:new("test_c4_queue")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	-- create an instance of the queue
	local c4Queue = c4_queue:new()

	c4Queue:SetMaxSize(3)
	assert(c4Queue:MaxSize() == 3, "_maxSize is not equal to '3' it is: " .. c4Queue:MaxSize())

	c4Queue:push("Item #1 in Queue")
	c4Queue:push("Item #2 in Queue")
	c4Queue:push("Item #3 in Queue")
	c4Queue:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4Queue:Size() == 1, "queue size is not equal to '1' it is: " .. c4Queue:Size())

	print (c4Queue)

	-- Test inheritance overriding OnMaxSizeReached
	-- Create a new class
	c4_queue_new = inheritsFrom(c4_queue)

	-- override construct()
	function c4_queue_new:construct()
		self.superClass():construct()  -- call base class
		self._maxSizeOption = 1

		local mt = getmetatable(self)
		if (mt ~= nil) then
			mt.__tostring = self.__tostring
		end
	end

	-- override OnMaxSizeReached()
	function c4_queue_new:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,

		-- Option 1: Do Nothing, new item is not added to queue
		if (self._maxSizeOption == 1) then
			
			LogInfo("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)
		-- Option 2: pop value, and push new value on queue
		elseif(self._maxSizeOption == 2) then
			LogInfo("Max Size Reached - pop value, and push new value on queue (option 2)")
			self:pop()
			return (true)
		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(self._maxSizeOption == 3) then
			LogInfo("Max Size Reached - clear queue and DO NOT push new value onto queue")
			self:clear()
			return (false)
		end
	end

	-- create an instance of the new queue
	local c4QueueNew = c4_queue_new:new()
	c4QueueNew:SetMaxSize(3)
	c4QueueNew:push("Item #1 in Queue")
	c4QueueNew:push("Item #2 in Queue")
	c4QueueNew:push("Item #3 in Queue")
	c4QueueNew:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4QueueNew:Size() == 3, "queue size is not equal to '3' it is: " .. c4QueueNew:Size())

	print(c4QueueNew)
	print ("done...")
end end)
package.preload['common.c4_device_connection_base'] = (function (...)
--[[=============================================================================
    DeviceConnectionBase Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_common"
require "lib.c4_object"
require "lib.c4_log"
require "lib.c4_timer"
require "lib.c4_queue"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_device_connection_base = "2015.03.02"
end

COMMAND_QUEUE_SIZE = 100
DEFAULT_COMMAND_DELAY_INTERVAL = 100            -- Don't send consecutive commands faster than this many milliseconds
DEFAULT_COMMAND_RESPONSE_INTERVAL = 3           -- If we haven't received and ACK after this many seconds, try again
DEFAULT_RETRY_COUNT_MAX = 3

function ON_DRIVER_EARLY_INIT.c4_device_connection_base()
	gReceiveBuffer = ""
	gIsUrlConnected = false
end

DeviceConnectionBase = inheritsFrom(nil)

function DeviceConnectionBase:construct()

	self._IsConnected = false
	self._SendTimer = nil
	self._WaitResponseTimer = nil
	self._CommandQueue = nil
	self._Priority1CommandQueue = nil
	self._Priority2CommandQueue = nil
	self._LastCommand = nil
	self._ExpectAck = false
	self._CommandRetryCount = 0
	self._RetryCountMax = DEFAULT_RETRY_COUNT_MAX

	self._SendCommandDelayMS = DEFAULT_COMMAND_DELAY_INTERVAL
	self._CommandResponseWaitS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	-- Polling
	self._PollingInterval = 0
	self._PollingUnits = "SECONDS"
	self._PollingTimer = nil
end

function DeviceConnectionBase:Initialize(ExpectAck, CommandDelayInterval, CommandResponseInterval, CallbackParam)

	if (ExpectAck ~= nil) then
		self._ExpectAck = ExpectAck
	end

	if (CommandDelayInterval ~= nil) then
		self._SendCommandDelayMS = CommandDelayInterval
	end

	if (CommandResponseInterval ~= nil) then
		self._CommandResponseWaitS = CommandResponseInterval
	end

	self._CommandQueue = c4_queue:new()
	self._CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._CommandQueue:SetName("Command Queue")

	self._Priority1CommandQueue = c4_queue:new()
	self._Priority1CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority1CommandQueue:SetName("P1 Queue")

	self._Priority2CommandQueue = c4_queue:new()
	self._Priority2CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority2CommandQueue:SetName("P2 Queue")

	-- usually only one of these timers will be used, but it's pretty low overhead to instantiate both of them
	self._SendTimer = c4_timer:new("SendCommand", self._SendCommandDelayMS, "MILLISECONDS", DeviceConnectionBase.OnSendTimeExpired, false, CallbackParam)
	self._WaitResponseTimer = c4_timer:new("WaitResponse", self._CommandResponseWaitS, "SECONDS", DeviceConnectionBase.OnWaitTimeExpired, false, CallbackParam)
end

function DeviceConnectionBase:InitPolling(PollingInterval, PollingUnits, CallbackParam)
	LogFatal("DeviceConnectionBase:InitPolling()")
	if (PollingInterval ~= nil) then
		self._PollingInterval = PollingInterval
	end
	
	self._PollingUnits = PollingUnits or self._PollingUnits

	LogFatal("self._PollingInterval: %s, self._PollingUnits: %s", tostring(self._PollingInterval), tostring(self._PollingUnits))
	
	-- create polling timer
	self._PollingTimer = c4_timer:new("Polling", self._PollingInterval, self._PollingUnits, DeviceConnectionBase.OnPollingTimerExpired, false, CallbackParam)
end

function DeviceConnectionBase:StartPolling(interval, units)
	LogFatal("DeviceConnectionBase:StartPolling()")
	LogFatal("self._PollingTimer: %s", tostring(self._PollingTimer))
	
	if (self._PollingTimer ~= nil) then
		self._PollingTimer:KillTimer()
		
		local timer_units = units or self._PollingTimer._units
		local timer_interval = interval or self._PollingInterval

		self._PollingTimer:StartTimer(timer_interval, timer_units)
	end
end

function DeviceConnectionBase:StopPolling()
	LogFatal("DeviceConnectionBase:StopPolling()")
	self._PollingTimer:KillTimer()
end

function DeviceConnectionBase:SetExpectACK(ExpectACK)
	self._ExpectAck = ExpectACK
end

function DeviceConnectionBase:SetCommandDelayInterval(DelayInterval)
	self._SendCommandDelayMS = DelayInterval
end

function DeviceConnectionBase:SetResponseWaitInterval(WaitInterval)
	self._CommandResponseWaitS = WaitInterval
end

function DeviceConnectionBase:ReceivedFromCom(sData)

	gReceiveBuffer = gReceiveBuffer .. sData
	LogTrace("ReceivedFromCom  ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)

	message = self:GetMessage()
	while (message ~= nil and message ~= "") do
		status, err = pcall(HandleMessage, message)
		if (status) then
			message = self:GetMessage()
		else
			LogError("LUA_ERROR: " .. err)
			message = ""
			gReceiveBuffer = ""
		end
	end
end

function DeviceConnectionBase:SetConnection(IsConnected, method)
	self._IsConnected = IsConnected
	gControlMethod = method
end

function DeviceConnectionBase:ControlMethod()
	-- Override in derived class
	print("WARNING: Need to override ControlMethod - should never be called")
	
	return ""
end

function DeviceConnectionBase:StartCommandTimer(...)
	local value = select(1, ...)
	local units = select(2, ...)
	local command_name = select(3, ...) or ""

	self._WaitResponseTimer:KillTimer()
	self._SendTimer:KillTimer()

	if (self._ExpectAck) then
		-- expecting an ACK set the Response Wait timer
		local timer_units = units or self._WaitResponseTimer._units
		local timer_interval = value or self._CommandResponseWaitS

		self._WaitResponseTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting wait Timer:  %d", self._WaitResponseTimer._timerID) .. " for " .. command_name)
	else
		-- no ACK expected, just wait the designated amount of time and send another command
		local timer_units = units or self._SendTimer._units
		local timer_interval = value or self._SendCommandDelayMS

		self._SendTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting Send Timer:  %d for %s (timer_interval = %d, timer_units = %s)", self._SendTimer._timerID, command_name, timer_interval, timer_units))
	end
end

-- Note the '.' instead of the ':'
function DeviceConnectionBase.OnSendTimeExpired(Instance)
	LogTrace("Send Timer expired")
	Instance._SendTimer:KillTimer()

	local tCommand = Instance._LastCommand
	if (tCommand.command_name ~= nil) then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand.command_name))
	elseif (type(tCommand) == "string") then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand))
	end
	
	Instance._LastCommand = nil
	Instance:SendNextCommand()

	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase.OnWaitTimeExpired(Instance)
	LogTrace("Wait Timer expired")
	Instance._WaitResponseTimer:KillTimer()
	Instance._CommandRetryCount = Instance._CommandRetryCount + 1

	if (Instance._CommandRetryCount >= Instance._RetryCountMax) then
		-- To many retries, pop the current command and try the next one
		Instance._CommandRetryCount = 0
		Instance:SendNextCommand()
	else
		Instance:SendLastCommand()
	end
end

function DeviceConnectionBase.OnPollingTimerExpired(Instance)
	LogTrace("Polling Timer expired")
	Instance._PollingTimer:KillTimer()

	OnPollingTimerExpired()
	
	Instance._PollingTimer:StartTimer(Instance._PollingInterval)
end

function DeviceConnectionBase:HandleACK()
	self._LastCommand = nil
	
	self._WaitResponseTimer:KillTimer()
	self._CommandRetryCount = 0
	self:SendNextCommand()
end

function DeviceConnectionBase:QueueEmpty()
	return (self._CommandQueue:empty() and self._Priority1CommandQueue:empty() and self._Priority2CommandQueue:empty())
end

function DeviceConnectionBase:QueueCommand(sCommand, ...)
--	LogTrace("QueueCommand(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority1Command(sCommand, ...)
	--LogTrace("QueuePriority1Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority1CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority2Command(sCommand, ...)
	LogTrace("QueuePriority2Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority2CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:SendNextCommand()
	LogTrace("DeviceConnectionBase:SendNextCommand")

	local tCommand = nil
	if (not self._Priority1CommandQueue:empty()) then
		tCommand = self._Priority1CommandQueue:pop()
		LogTrace(tostring(gCon._Priority1CommandQueue))
	elseif (not self._Priority2CommandQueue:empty()) then
		tCommand = self._Priority2CommandQueue:pop()
		LogTrace(tostring(gCon._Priority2CommandQueue))
	elseif (not self._CommandQueue:empty()) then
		tCommand = self._CommandQueue:pop()
		LogTrace(tostring(gCon._CommandQueue))
	end
	
	if (tCommand ~= nil) then
		self._LastCommand = tCommand
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendLastCommand()
--	LogTrace("DeviceConnectionBase:SendLastCommand")

	local tCommand = self._LastCommand
	if (tCommand ~= nil) then
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendCommand()
	-- Dummy routine.  Override in derived class
	print("Need to override SendCommand - should never be called")
end


function DeviceConnectionBase:GetMessage()
	
	-- Brain dead version of this routine. Just return the current receive buffer.
	-- It's very likely that a GetMessage() function will need to be created
	if (GetMessage ~= nil and type(GetMessage) == "function") then
		return GetMessage()
	else
		local ComMessage = gReceiveBuffer
		gReceiveBuffer = ""

		return ComMessage
	end
end

--[[=============================================================================
    Other Connection Functions
===============================================================================]]

function ReceivedFromSerial(idBinding, sData)
	if (gCon.ReceivedFromSerial == nil) then return end --serial is bound but not the current control method
	gCon:ReceivedFromSerial(idBinding, sData)
end

function ReceivedFromNetwork(idBinding, nPort, sData)
	gCon:ReceivedFromNetwork(idBinding, nPort, sData)
end

function OnServerDataIn(nHandle, strData)
--	LogTrace("Received Data on Handle: " .. nHandle .. ": " .. strData)
--	LogTrace("Data Is: %s", HexToString(strData))
	gCon:ReceivedFromNetworkServer(nHandle, strData)
end


--[[=============================================================================
    The ReceivedAsync function is called in response to 'url_get_request'. 
    The ticketId is the number returned from the request.
===============================================================================]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

--	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
--	LogTrace(tHeaders)

	gCon:ReceivedAsync(ticketId, strData, responseCode, tHeaders)
end
	
--[[=============================================================================
    OnBindingChanged(idBinding, class, bIsBound)
  
    Description:
    Function called by Director when a binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnBindingChanged(idBinding, class, bIsBound)
	
	LogTrace("OnBindingChanged(): idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	if (idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
		SetControlMethod()
		OnSerialConnectionChanged(idBinding, class, bIsBound)
	elseif (idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
		SetControlMethod()
		OnIRConnectionChanged(idBinding, class, bIsBound)
	end
end

--[[=============================================================================
    OnNetworkBindingChanged(idBinding, bIsBound)
  
    Description:
    Function called by Director when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	LogTrace('OnNetworkBindingChanged(): idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))

	gIsNetworkConnected = bIsBound
	SetControlMethod()
	OnNetworkConnectionChanged(idBinding, bIsBound)
	if (bIsBound) then
		-- Start a special instance of reconnect timer to eventually do NetConnect if not done automatically
		gCon._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval) 	
	end	
end

--[[=============================================================================
    OnConnectionStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. tostring(nPort) .. ")]: " .. sStatus)
     if (nPort ~= NETWORK_PORT) then
	   LogWarn("OnConnectionStatusChanged, nPort(" .. nPort .. ") is not valid.  ***************This is an unused port that the projector is opening and closing.  This has nothing to do with the driver or it’s functionality.  Please ignore this port when troubleshooting. ************************")  
	   return
	end
	
	local isOnline = false

	gNetworkStatus = sStatus	
	if (sStatus == "ONLINE") then
		isOnline = true
	end

	gCon:SetOnlineStatus(isOnline)
	OnNetworkStatusChanged(idBinding, nPort, sStatus)
end

--[[=============================================================================
    SetControlMethod()
  
    Description:
    Sets the control method type for the drivers internal infrastructure
  
    Parameters:
    None
  
    Returns:
    The type of control method for the drivers connection(i.e. Network, Serial,
    IR, or (none))
===============================================================================]]
function SetControlMethod()
	if (gIsNetworkConnected == false) and (gCon._NetworkReconnectTimer ~= nil) then
		--housekeeping when changing from network control to serial or IR control
		gCon._NetworkReconnectTimer:KillTimer() 
	end


	if( gIsNetworkServerConnected) then
		-- connect to NetworkServer communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "NetworkServer") then
			gCon = NetworkServerConnectionBase:new()
			gCon:Initialize()
		end
		gCon:SetConnection(true, "NetworkServer")
	elseif (gIsNetworkConnected) then
		-- connect to Network communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Network") then
			gCon = NetworkConnectionBase:new(NETWORK_BINDING_ID, NETWORK_PORT)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "Network")
	elseif (gIsUrlConnected) then
		-- connect to URL communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "URL") then
			gCon = UrlConnectionBase:new()
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "URL")
	elseif (gIsSerialConnected) then
		-- connect to Serial communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Serial") then
			gCon = SerialConnectionBase:new(SERIAL_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
			gCon:InitPolling(tonumber(gPollingTimerInterval), "MINUTES", gCon)
		end
		gCon:SetConnection(true, "Serial")
	elseif (gIsIRConnected) then
		-- connect to IR communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "IR") then
			gCon = IRConnectionBase:new(IR_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "IR")
	else
		if (gCon ~= nil) then
			gCon:SetConnection(false, "(none)")
		end
		-- gCon = nil
	end

	gCon._CommandQueue:clear()
	gCon._Priority1CommandQueue:clear()
	gCon._Priority2CommandQueue:clear()
end

--[[=============================================================================
    ValidateControlMethod(controlMethod)
  
    Description:
    Identifies whether the specified control method has a valid connection
  
    Parameters:
    controlMethod(string) - The communication we are validating against
                            Valid types are (Network, Serial, and IR)
  
    Returns:
    true if the controlMethod specified has been connected, false otherwise.
===============================================================================]]
function ValidateControlMethod(controlMethod)
	local isValid = false

	if (controlMethod == "Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (controlMethod == "URL") and (gIsUrlConnected) then
		isValid = true
	elseif (controlMethod == "Serial") and (gIsSerialConnected) then
		isValid = true
	elseif (controlMethod == "IR") and (gIsIRConnected) then
		isValid = true
	end

	return isValid
end
 end)
package.preload['common.c4_network_connection'] = (function (...)
--[[=============================================================================
    Base for a network connection driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"
require "lib.c4_log"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_network_connection = "2015.03.02"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30
DEFAULT_RECONNECT_INTERVAL_SECONDS = 5

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS
gNetworkReconnectInterval = DEFAULT_RECONNECT_INTERVAL_SECONDS

NetworkConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkConnectionBase:construct(BindingID, Port)
	self.superClass():construct()

	self._BindingID = BindingID
	self._Port = Port
	self._LastCheckin = 0
	self._IsOnline = false
	self._KeepAliveTimer = nil
end

function NetworkConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetConBase:Initialize")
	gControlMethod = "Network"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	self._KeepAliveTimer = c4_timer:new("PollingTimer", gNetworkKeepAliveInterval, "SECONDS", NetworkConnectionBase.OnKeepAliveTimerExpired, false, self)
	self._NetworkReconnectTimer = c4_timer:new("NetworkReconnectTimer", gNetworkReconnectInterval, "SECONDS", NetworkConnectionBase.OnNetworkReconnectTimerExpired, false, self)
end

function NetworkConnectionBase:ControlMethod()
	return "Network"
end

function NetworkConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)

			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end

function NetworkConnectionBase:ReceivedFromNetwork(idBinding, nPort, sData)
	self._LastCheckin = 0
	self:ReceivedFromCom(sData)
end

function NetworkConnectionBase:CheckNetworkConnectionStatus()
	if (self._IsConnected and (not self._IsOnline)) then
		LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(self._BindingID, self._Port)
		--C4:NetConnect(self._BindingID, self._Port)
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	end
end

function NetworkConnectionBase.OnKeepAliveTimerExpired(Instance)
	Instance._LastCheckin = Instance._LastCheckin + 1

	if(Instance._LastCheckin == 3) then
		LogWarn("Failed to receive poll responses... initiating network recovery mode...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		return
	elseif(Instance._LastCheckin > 4) then	
		Instance._LastCheckin = 4
	end

	if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function" and Instance._IsOnline) then
		SendKeepAlivePollingCommand()
	end

	Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
end

function NetworkConnectionBase.OnNetworkReconnectTimerExpired(Instance)
	if (Instance._IsConnected) then
		LogWarn("OnNetworkReconnectTimerExpired: Attempting to reactivate network connection...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		C4:NetConnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	else
		LogWarn("Cannot attempt to reactivate, the network connection is not bound")
	end
end

function NetworkConnectionBase:SetOnlineStatus(IsOnline)
	self._IsOnline = IsOnline

	if(IsOnline) then
		self._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
		self._NetworkReconnectTimer:KillTimer()
		self._LastCheckin = 0
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "true")
		end

		self:SendNextCommand()
	else
		self._KeepAliveTimer:KillTimer()
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "false")
		end
	end
end

function ON_DRIVER_LATEINIT.c4_network_connection()
	-- Ensure existing connection is taken into consideration (useful on Driver Update)
	if (gCon ~= nil) then
		local tmp = C4:GetBindingAddress(gCon._BindingID)
		if (tmp ~= nil and string.len(tmp) > 0) then OnNetworkBindingChanged(gCon._BindingID, true) end
	end
end
 end)
package.preload['common.c4_notify'] = (function (...)
--[[=============================================================================
    Notification Functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_notify = "2015.03.02"
end

--[[=============================================================================
    SendNotify(notifyText, tParams, bindingID)

    Description
    Forwards a notification to the proxy with a list of parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    tParams(table)     - Table of key value pairs that hold the the parameters
                         and their values used in the proxy function
    bindingID(int)     - The requests binding id

    Returns
    Nothing
===============================================================================]]
function SendNotify(notifyText, tParams, bindingID)
	C4:SendToProxy(bindingID, notifyText, tParams, "NOTIFY")
end

--[[=============================================================================
    SendSimpleNotify(notifyText, ...)

    Description
    Forwards a notification to the proxy with no parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    bindingID(int)     - Optional parameter containing the requests binding id,
                         if not specified then the DEFAULT_PROXY_ID is given.

    Returns
    Nothing
===============================================================================]]
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end end)
package.preload['common.c4_command'] = (function (...)
--[[=============================================================================
    Functions for handling and executing commands and actions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_command = "2015.03.02"
end

--[[=============================================================================
    ExecuteCommand(sCommand, tParams)

    Description
    Function called by Director when a command is received for this DriverWorks
    driver. This includes commands created in Composer programming.

    Parameters
    sCommand(string) - Command to be sent
    tParams(table)   - Lua table of parameters for the sent command

    Returns
    Nothing
===============================================================================]]
function ExecuteCommand(sCommand, tParams)
	LogTrace("ExecuteCommand(" .. sCommand .. ")")
	LogInfo(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")
	local status, err

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		status, err = pcall(EX_CMD[sCommand], tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		status, err = pcall(EX_CMD[trimmedCommand], tParams)
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
		status = true
	else
		LogInfo("ExecuteCommand: Unhandled command = " .. sCommand)
		status = true
	end
	
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    EX_CMD.LUA_ACTION(tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    tParams(table) - Lua table of parameters for the command option

    Returns
    Nothing
===============================================================================]]
function EX_CMD.LUA_ACTION(tParams)
	if (tParams ~= nil) then
		for cmd, cmdv in pairs(tParams) do
			if (cmd == "ACTION" and cmdv ~= nil) then
				local status, err = pcall(LUA_ACTION[cmdv], tParams)
				if (not status) then
					LogError("LUA_ERROR: " .. err)
				end
				break
			end
		end
	end
end

--[[=============================================================================
    ReceivedFromProxy(idBinding, sCommand, tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    idBinding(int)   - Binding ID of the proxy that sent a BindMessage to the
                       DriverWorks driver.
    sCommand(string) - Command that was sent
    tParams(table)   - Lua table of received command parameters

    Returns
    Nothing
===============================================================================]]
function ReceivedFromProxy(idBinding, sCommand, tParams)

	if (sCommand ~= nil) then

		-- initial table variable if nil
		if (tParams == nil) then
			tParams = {}
		end
		
		LogTrace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function PRX_CMD." .. sCommand .. "()")
		LogInfo(tParams)

		if ((PRX_CMD[sCommand]) ~= nil) then
			local status, err = pcall(PRX_CMD[sCommand], idBinding, tParams)
			if (not status) then
				LogError("LUA_ERROR: " .. err)
			end
		else
			LogInfo("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end end)
package.preload['common.c4_property'] = (function (...)
--[[=============================================================================
    Function for changing properties

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_property = "2015.03.02"
end

--[[=============================================================================
    OnPropertyChanged(sProperty)

    Description
    Function called by Director when a property changes value. The value of the
    property that has changed can be found with: Properties[sName]. Note that
    OnPropertyChanged is not called when the Property has been changed by the
    driver calling the UpdateProperty command, only when the Property is changed
    by the user from the Properties Page. This function is called by Director
    when a property changes value.

    Parameters
    sProperty(string) - Name of property that has changed.

    Returns
    Nothing
===============================================================================]]
function OnPropertyChanged(sProperty)
	local propertyValue = Properties[sProperty]

	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	end

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	local status = true
	local err = ""

	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[sProperty], propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[trimmedProperty], propertyValue)
	end

	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)
	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end
 end)
package.preload['common.c4_init'] = (function (...)
--[[=============================================================================
    Initial driver initialization and destruction functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_property"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_init = "2015.03.02"
end

--[[=============================================================================
    OnDriverInit()

    Description
    Invoked by director when a driver is loaded. This API is provided for the
    driver developer to contain all of the driver objects that will require
    initialization.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverInit()
	gInitializingDriver = true
	C4:ErrorLog("INIT_CODE: OnDriverInit()")

	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_EARLY_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property
	-- global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		C4:ErrorLog("INIT_CODE: Calling OnPropertyChanged - " .. k .. ": " .. v)
		local status, err = pcall(OnPropertyChanged, k)
		if (not status) then
			C4:ErrorLog("LUA_ERROR: " .. err)
		end
	end

	gInitializingDriver = false
end

--[[=============================================================================
    OnDriverLateInit()

    Description
    Invoked by director after all drivers in the project have been loaded. This
    API is provided for the driver developer to contain all of the driver
    objects that will require initialization after all drivers in the project
    have been loaded.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[=============================================================================
    OnDriverDestroyed()
    Function called by Director when a driver is removed. Release things this
    driver has allocated such as timers.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	
	-- Call all ON_DRIVER_DESTROYED functions
	for k, v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end end)
package.preload['lib.c4_timer'] = (function (...)
--[[=============================================================================
    c4_timer Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_timer = "2015.03.02"
end

c4_timer = inheritsFrom(nil)

function c4_timer:construct(name, interval, units, Callback, repeating, CallbackParam)
	self._name = name
	self._timerID = TimerLibGetNextTimerID()
	self._interval = interval
	self._units = units
	self._repeating = repeating or false
	self._Callback = Callback
	self._CallbackParam = CallbackParam or ""
	self._id = 0

	gTimerLibTimers[self._timerID] = self
	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("Created timer " .. self._name)
	end
end

function c4_timer:StartTimer(...)
	c4_timer:KillTimer()

	-- optional parameters (interval, units, repeating)
	if ... then
		local interval = select(1, ...)
		local units = select(2, ...)
		local repeating = select(3, ...)

		self._interval = interval or self._interval
		self._units = units or self._units
		self._repeating = repeating or self._repeating
	end

	if (tonumber(self._interval) > 0) then
		if (LOG ~= nil and type(LOG) == "table") then
			LogTrace("Starting Timer: " .. self._name)
		end

		self._id = C4:AddTimer(self._interval, self._units, self._repeating)
	end
end

function c4_timer:KillTimer()
	if (self._id) then
		self._id = C4:KillTimer(self._id)
	end
end

function c4_timer:TimerStarted()
	return (self._id ~= 0)
end

function c4_timer:TimerStopped()
	return (self._id == 0)
end

function c4_timer:GetTimerInterval()
	return (self._interval)
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.c4_timer()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.c4_timer()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[=============================================================================
    OnTimerExpired(idTimer)

    Description:
    Function called by Director when the specified Control4 timer expires.

    Parameters:
    idTimer(string) - Timer ID of expired timer.
===============================================================================]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._CallbackParam)
			end
		end
	end
end

--[[=============================================================================
    CreateTimer(name, interval, units, callback, repeating, callbackParam)

    Description: 
    Creates a named timer with the given attributes

    Parameters:
    name(string)       - The name of the timer being created
    interval(int)      - The amount of the given time between calls to the
                         timers callback function
    units(string)      - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    callback(string)   - The function to call when the timer expires
    repeating(bool)    - Parameter indicating whether the timer should be
                         called repeatedly until cancelled
    callbackParam(...) - Parameters to be passed to the callback function

    Returns:
    A handle to the timer
===============================================================================]]
function CreateTimer(name, interval, units, callback, repeating, callbackParam)
	timer = c4_timer:new(name, interval, units, callback, repeating, callbackParam)
	return timer
end

--[[=============================================================================
    StartTimer(handle, ...)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer)   - Handle to a created timer object
    interval(int)   - The amount of the given time between calls to the
                      timers callback function
    units(string)   - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    repeating(bool) - Parameter indicating whether the timer should be
                      called repeatedly until cancelled

    Returns:
    None
===============================================================================]]
function StartTimer(handle, ...)
	handle:StartTimer(...)
end

--[[=============================================================================
    KillTimer(handle)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    None
===============================================================================]]
function KillTimer(handle)
	handle:KillTimer()
end

--[[=============================================================================
    TimerStarted(handle)

    Description: 
    Identifies whether a timer has been started or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been started, or false otherwise
===============================================================================]]
function TimerStarted(handle)
	return handle:TimerStarted()
end

--[[=============================================================================
    TimerStopped(handle)

    Description: 
    Identifies whether a timer has been stopped or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been stopped, or false otherwise
===============================================================================]]
function TimerStopped(handle)
	return handle:TimerStopped()
end

--[[=============================================================================
    GetTimerInterval(handle)

    Description: 
    Gets the interval setting of the given timer

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns the interval setting of the given timer
===============================================================================]]
function GetTimerInterval(handle)
	return handle:GetTimerInterval()
end

--[[=============================================================================
    c4_timer Unit Tests
===============================================================================]]
function __test_c4_timer()
	require "test.C4Virtual"
	require "lib.c4_log"
	require "common.c4_init"

	OnDriverInit()

	local LOG = c4_log:new("test_c4_timer")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	function OnTestTimerExpired()
		c4Timer:KillTimer()
	end

	-- create an instance of the timer
	c4Timer = c4_timer:new("Test", 45, "MINUTES", OnTestTimerExpired)

	assert(c4Timer._id == 0, "_id is not equal to '0' it is: " .. c4Timer._id)
	c4Timer:StartTimer()
	assert(c4Timer._id == 10001, "_id is not equal to '10001' it is: " .. c4Timer._id)
	assert(c4Timer:TimerStarted() == true, "TimerStarted is not equal to true it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == false, "TimerStopped is not equal to false it is: " .. tostring(c4Timer:TimerStopped()))
	OnTimerExpired(c4Timer._id)
	assert(c4Timer:TimerStarted() == false, "TimerStarted is not equal to false it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == true, "TimerStopped is not equal to true it is: " .. tostring(c4Timer:TimerStopped()))
end end)
package.preload['lib.c4_object'] = (function (...)
--[[=============================================================================
    c4_object Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_object = "2015.03.02"
end

function inheritsFrom( baseClass )
	local new_class = {}
	local class_mt = { __index = new_class }

	function new_class:create(...)
		local newinst = {}

		setmetatable( newinst, class_mt )

		-- Call the constructor when we create this class
		if newinst.construct then
			-- Allow returning a different obj than self. This allows for readonly tables etc...
			newinst = newinst:construct(...) or newinst
		end

		return newinst
	end

	if nil ~= baseClass then
		setmetatable( new_class, { __index = baseClass } )
	end

	--[[=============================================================================
	    Implementation of additional OO properties starts here 
	===============================================================================]]
	
	-- Return the class object of the instance
	function new_class:class()
		return new_class
	end

	--[[=============================================================================
	    Return the super class object of the instance. 

	    Note Calling methods on the base class itself will modify
	    the base table's static properties. In order to have call
	    the base class' methods and have them modify the current object
	    use super() or superAsSelf(). 
	===============================================================================]]
	function new_class:superClass()
		return baseClass
	end

	--[[=============================================================================
	    Returns a table that allows calling of the base class's method
	    while maintaining the objects state as the modified state of the base
	    class' methods. For example consider the following statements (order matters):

	    -- The child sees the parents property if the child hasn't overriden the property
	       obj:superClass().id = "parent"
	       obj.id == "parent" -- true

	    -- Setting the property on the child overrides (hides) the parents property
	       obj.id = "child"
	       obj.id == "child" -- true

	    -- The super() method pass
	       obj:super().id == "parent" -- true
	       obj:super().id = "child"
	       obj:super().id == "parent" -- still true
	       obj.id == "child" -- still true
	===============================================================================]]
	function new_class:super()
		local holder = {}

		holder.child = self
		holder.parent = baseClass

		local mt = {}
		mt.__index = function(table, index)
			if table.parent[index] then
				return table.parent[index]
			else 
				return table.child[index]
			end
		end

		-- Only set the new values to the child.
		mt.__newindex = function(table, key, value)
			table.child[key] = value
		end

		mt.__tostring = function(table)
			return tostring(table.child)
		end

		setmetatable(holder, mt)
		return holder
	end

	new_class.new = new_class.create

	--[[=============================================================================
	    Return true if the caller is an instance of theClass
	===============================================================================]]
	function new_class:isa( theClass )
		local b_isa = false
		local cur_class = new_class

		while ( nil ~= cur_class ) and ( false == b_isa ) do
			if cur_class == theClass then
				b_isa = true
			else
				cur_class = cur_class:superClass()
			end
		end

		return b_isa
	end

	return new_class
end

--[[=============================================================================
    Inheritance unit tests
===============================================================================]]
function __test_inheritance()
	local b = inheritsFrom(nil)

	b.construct = function(self, msg)
		self._msg = msg
	end

	local t = inheritsFrom(b)
	t.construct = function(self, msg)
		self:super():construct(msg)
	end

	t1 = t:new("t1")
	t2 = t:new("t2")
	assert(t1._msg == "t1", "t1 message is not equal to 't1' it''s: " .. t1._msg)
	assert(t2._msg == "t2", "t2 message is not equal to 't2' it''s: " .. t2._msg)    
	assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");
	assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

	t1:superClass().id = "parent"
	assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

	-- Setting the property on the child overrides (hides) the parents property
	t1.id = "child"
	assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

	-- The super() method maintains the self pointer to the child and not to the base
	assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
	t1:super().id = "child1"
	assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
	assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")
end end)
package.preload['lib.c4_log'] = (function (...)
--[[=============================================================================
    c4_log Class

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_log = "2015.03.02"
end

c4_log = inheritsFrom(nil)

function c4_log:construct(logName)
	self._logLevel = tonumber(string.sub(Properties['Log Level'] or "", 1, 1)) or 5
	self._outputPrint = Properties['Log Mode']:find("Print") ~= nil
	self._outputC4Log = Properties['Log Mode']:find("Log") ~= nil
	self._logName = logName or ""

	-- make sure Property is up to date (no harm if absent)
	C4:UpdateProperty("Log Level", Properties['Log Level'])
end

function c4_log:SetLogLevel(level)
	self._logLevel = tonumber(string.sub(level or "", 1, 1)) or self._logLevel
end

function c4_log:LogLevel()
	return self._logLevel
end

function c4_log:OutputPrint(value)
	self._outputPrint = value
end

function c4_log:OutputC4Log(value)
	self._outputC4Log = value
end

function c4_log:SetLogName(logName)

	if (logName == nil or logName == "") then
		logName = ""
	else
		logName = logName .. ": "
	end

	self._logName = logName
end

function c4_log:LogName()
	return self._logName
end

function c4_log:Enabled()
	return (self._outputPrint or self._outputC4Log)
end

function c4_log:PrintEnabled()
	return self._outputPrint
end

function c4_log:C4LogEnabled()
	return self._outputC4Log
end

function c4_log:CreateTableText(tValue, tableText)
	tableText = tableText or ""

	if (type(tValue) == "table") then

		tableText = tableText .. "{"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ","
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableText(v, tableText)
				tableText = tableText .. ","
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ","
			end
		end

		tableText = tableText .. "}"
	end

	return tableText
end

function InsertIndent(indentLevel)
	local indentStr = ""

	for i=1, indentLevel do
		indentStr = indentStr .. "\t"
	end

	return indentStr
end

function c4_log:CreateTableTextFormatted(tValue, tableText, indentLevel)
	tableText = tableText or ""
	indentLevel = indentLevel or 0

	if (type(tValue) == "table") then

		indentLevel = indentLevel + 1
		tableText = tableText .. "{\n"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. InsertIndent(indentLevel) .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. InsertIndent(indentLevel) .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ",\n"
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',\n"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableTextFormatted(v, tableText, indentLevel)
				tableText = tableText .. ",\n"
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ",\n"
			end
		end

		indentLevel = indentLevel - 1
		tableText = tableText .. InsertIndent(indentLevel) .. "}"
	end

	return tableText
end

MAX_TABLE_LEVELS = 10
function c4_log:PrintTable(tValue, tableText, sIndent, level)
	tableText = tableText or ""
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				if (tableText == "") then
					tableText = sIndent .. tostring(k) .. ":  " .. tostring(v)
					if (sIndent == ".   ") then sIndent = "    " end
				else
					tableText = tableText .. "\n" .. sIndent .. tostring(k) .. ":  " .. tostring(v)
				end
				if (type(v) == "table") then
					tableText = self:PrintTable(v, tableText, sIndent .. "   ", level)
				end
			end
		else
			tableText = tableText .. "\n" .. sIndent .. tostring(tValue)
		end
	end
	
	return tableText
end

function c4_log:LogTable(tValue, sIndent, level)
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				C4:ErrorLog(self._logName .. sIndent .. tostring(k) .. ":  " .. tostring(v))
				if (type(v) == "table") then
					self:LogTable(v, sIndent .. "   ", level)
				end
			end
		else
			C4:ErrorLog(self._logName .. sIndent .. tValue)
		end
	end
end

function c4_log:Print(logLevel, sLogText)

	if (self._logLevel >= logLevel) then
		if (type(sLogText) == "table") then
			if (self._outputPrint) then
				print (self:PrintTable(sLogText, tableText, ".   ", 0))
			end

			if (self._outputC4Log) then
				self:LogTable(sLogText, "   ", 0)
			end

			return
		end

		if (self._outputPrint) then
			print (sLogText)
		end

		if (self._outputC4Log) then
			C4:ErrorLog(self._logName .. tostring(sLogText))
		end
	end
end

function c4_log:Fatal(sLogText, ...)
	self:LogOutput(0, sLogText, ...)
end

function c4_log:Error(sLogText, ...)
	self:LogOutput(1, sLogText, ...)
end

function c4_log:Warn(sLogText, ...)
	self:LogOutput(2, sLogText, ...)
end

function c4_log:Info(sLogText, ...)
	self:LogOutput(3, sLogText, ...)
end

function c4_log:Debug(sLogText, ...)
	self:LogOutput(4, sLogText, ...)
end

function c4_log:Trace(sLogText, ...)
	self:LogOutput(5, sLogText, ...)
end

function c4_log:LogOutput(level, sLogText, ...)
	if (LogEnabled()) then
		if (type(sLogText) == "string") then
			sLogText = string.format(sLogText, ...)
		end

		self:Print(level, sLogText)
	end
end

--[[=============================================================================
    c4_log wrapper functions
===============================================================================]]
function TryLog(level, sLogText, ...)
	LOG:LogOutput(level, sLogText, ...)
end

--[[=============================================================================
    SetLogLevel(level)

    Description: 
    Sets the desired log level to view

    Parameters:
    level(int) - The logging level to set the message to
                 0 = Fatal
                 1 = Error
                 2 = Warn
                 3 = Info
                 4 = Debug
                 5 = Trace

    Returns:
    None
===============================================================================]]
function SetLogLevel(level)
	LOG:SetLogLevel(level)
end

--[[=============================================================================
    LogLevel()

    Description: 
    Returns the currently set log level

    Parameters:
    None

    Returns:
    The current log level
        0 = Fatal
        1 = Error
        2 = Warn
        3 = Info
        4 = Debug
        5 = Trace
===============================================================================]]
function LogLevel()
	return LOG:LogLevel()
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputPrint(value)
	LOG:OutputPrint(value)
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages to file or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputC4Log(value)
	LOG:OutputC4Log(value)
end

--[[=============================================================================
    SetLogName(logName)

    Description: 
    Sets the name of the log file where the messages will be written to

    Parameters:
    logName(string) - Sets the name of the log to write messages to

    Returns:
    None
===============================================================================]]
function SetLogName(logName)
	LOG:SetLogName(logName)
end

--[[=============================================================================
    LogName(logName)

    Description: 
    Gets the name of the log file where the messages will be written to

    Parameters:
    None

    Returns:
    The value of the log file that has been set
===============================================================================]]
function LogName()
	return LOG:LogName()
end

--[[=============================================================================
    LogEnabled()

    Description: 
    Identifies if logging or print has been enabled

    Parameters:
    None

    Returns:
    true if either logging or print has been enabled, false otherwise
===============================================================================]]
function LogEnabled()
	return LOG:Enabled()
end

--[[=============================================================================
    PrintEnabled()

    Description: 
    Gets the state of print output

    Parameters:
    None

    Returns:
    true if print has been enabled, false otherwise
===============================================================================]]
function PrintEnabled()
	return LOG:PrintEnabled()
end

--[[=============================================================================
    C4LogEnabled()

    Description: 
    Gets the state of logging

    Parameters:
    None

    Returns:
    true if logging has been enabled, false otherwise
===============================================================================]]
function C4LogEnabled()
	return LOG:C4LogEnabled()
end

--[[=============================================================================
    LogFatal(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Fatal(0) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogFatal(sLogText, ...)
	local status, err = pcall(TryLog, 0, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogFatal failed: " .. err)
	end
end

--[[=============================================================================
    LogError(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Error(1) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogError(sLogText, ...)
	local status, err = pcall(TryLog, 1, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogError failed: " .. err)
	end
end

--[[=============================================================================
    LogWarn(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Warn(2) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogWarn(sLogText, ...)
	local status, err = pcall(TryLog, 2, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogWarn failed: " .. err)
	end
end

--[[=============================================================================
    LogInfo(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Info(3) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogInfo(sLogText, ...)
	local status, err = pcall(TryLog, 3, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogInfo failed: " .. err)
	end
end

--[[=============================================================================
    LogDebug(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Debug(4) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogDebug(sLogText, ...)
	local status, err = pcall(TryLog, 4, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogDebug failed: " .. err)
	end
end

--[[=============================================================================
    LogTrace(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Trace(5) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogTrace(sLogText, ...)
	local status, err = pcall(TryLog, 5, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogTrace failed: " .. err)
	end
end

function dbgPrint(buf)
	if (LOG:PrintEnabled()) then
		print (buf)
	end
end

function dbgHexdump(buf)
	hexdump(buf, dbgPrint)
end

--[[=============================================================================
    c4_log unit tests
===============================================================================]]
function __test_c4_log()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	assert(LOG:LogName() == "test_c4_log", "_logName is not equal to 'test_c4_log' it is: " .. LOG:LogName())

	-- Test setting log level
	LOG:SetLogLevel("2 - Warning")
	assert(LOG:LogLevel() == 2, "_logLevel is not equal to '2' it is: " .. LOG:LogLevel())

	LOG:SetLogLevel(3)
	assert(LOG:LogLevel() == 3, "_logLevel is not equal to '3' it is: " .. LOG:LogLevel())

	-- Test enabling logs
	LOG:OutputPrint(false)
	assert(LOG:PrintEnabled() == false, "_outputPrint is not equal to 'false' it is: " .. tostring(LOG:PrintEnabled()))

	LOG:OutputC4Log(true)
	assert(LOG:C4LogEnabled() == true, "_outputC4Log is not equal to 'true' it is: " .. tostring(LOG:C4LogEnabled()))

	LOG:SetLogLevel(4)
	LogTrace("***** This is a test *****")
end

function __test_CreatTableText()
	local tTest = {}

	tTest[1] = {}
	tTest[2] = {}
	tTest[3] = 30
	tTest[4] = "Forty"
	
	LogTrace("----- tText -----")
	LogTrace(tTest)

	local tTest2 = { One = {},
					 Two = {},
					 Three = 30,
					 Four = "Forty" }
	LogTrace("----- tText2 -----")
	LogTrace(tTest2)
	
	local tTest3 = { [1] = {},
					 [2] = {},
					 [3] = 30,
					 [4] = "Forty" }
	LogTrace("----- tText3 -----")
	LogTrace(tTest3)

	local tTest4 = { [1] = {},
					 Two = {},
					 [3] = 30,
					 [4] = "Forty",
					 Five = "Fifty" }
	LogTrace("----- tText4 -----")
	LogTrace(tTest4)

	local tableText = LOG:CreateTableText(tTest4)
	LogTrace("----- tableText -----")
	LogTrace(tableText)
	
	--local tNew = {[1] = {},[3] = 30,[4] = 'Forty',Five = 'Fifty',Two = {},}
	--LogTrace(tNew)
end

function __TestCreateTableTextFormatted()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	local tButtons = {
		Name = 'heat',
		Attributes = {},
		ChildNodes = {
			[1] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '51',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Pool Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'POOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[2] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '53',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Spa Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'SPAHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[3] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '54',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {Value = 'Pool Solar Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {}
					},
					[3] = {
						Value = 'SOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				}
			}
		}
	}

	print(LOG:CreateTableTextFormatted(tButtons))
end end)
package.preload['common.c4_common'] = (function (...)
--[[=============================================================================
    ON_INIT, Timer,s and Property management functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_log"
require "lib.c4_timer"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_common = "2015.03.02"
end

--[[=============================================================================
    Create and Initialize Logging
===============================================================================]]
function ON_DRIVER_EARLY_INIT.c4_common()
	-- Create a logger
	LOG = c4_log:new("Template_c4z Change Name")
end

function ON_DRIVER_INIT.c4_common()
	-- Create Log Timer
	gC4LogTimer = c4_timer:new("Log Timer", 45, "MINUTES", OnLogTimerExpired)
end

--[[=============================================================================
    Log timer callback function
===============================================================================]]
function OnLogTimerExpired()
	LogWarn("Turning Log Mode Off (timer expired)")
	gC4LogTimer:KillTimer()
	
	C4:UpdateProperty("Log Mode", "Off")
	OnPropertyChanged("Log Mode")
end

gForceLogging = false
function ON_PROPERTY_CHANGED.LogMode(propertyValue)
	gC4LogTimer:KillTimer()
	
	if (gForceLogging) then
		LOG:OutputPrint(true)
		LOG:OutputC4Log(true)
	else
		LOG:OutputPrint(propertyValue:find("Print") ~= nil)
		LOG:OutputC4Log(propertyValue:find("Log") ~= nil)
		if (propertyValue == "Off") then
			return
		end
		
		gC4LogTimer:StartTimer()
	end
end

function ON_PROPERTY_CHANGED.LogLevel(propertyValue)
	if (gForceLogging) then
		LOG:SetLogLevel("5 - Debug")
	else
		LOG:SetLogLevel(propertyValue)
	end
end

--[[=============================================================================
    Print Template Versions
===============================================================================]]
function TemplateVersion()
	print ("\nTemplate Versions")
	print ("-----------------------")
	for k, v in pairs(TEMPLATE_VERSION) do
		print (k .. " = " .. v)
	end
	
	print ("")
end end)
package.preload['common.c4_driver_declarations'] = (function (...)
--[[=============================================================================
    Driver Declarations used to call startup routines, teardown routines, and 
    other basic functions of the drivers operation

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Template Version Table
TEMPLATE_VERSION = {}
TEMPLATE_VERSION.c4_driver_declarations = "2015.03.02"

-- Command Handler Tables
EX_CMD = {}
PRX_CMD = {}
UI_REQ = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}

--[[=============================================================================
    Tables of functions
    The following tables are function containers that are called within the
    following functions:

    OnDriverInit()
        First calls all functions contained within ON_DRIVER_EARLY_INIT table
        then calls all functions contained within ON_DRIVER_INIT table

    OnDriverLateInit()
        Calls all functions contained within ON_DRIVER_LATEINIT table

    OnDriverDestroyed()
        Calls all functions contained within ON_DRIVER_DESTROYED table

    OnPropertyChanged()
        Calls all functions contained within ON_PROPERTY_CHANGED table
===============================================================================]]
ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001 end)
--[[=============================================================================
    Basic Template for PROJECTOR Driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]------------
require "common.c4_driver_declarations"
require "common.c4_common"
require "common.c4_init"
require "common.c4_property"
require "common.c4_command"
require "common.c4_notify"
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_utils"
require "lib.c4_timer"
require "actions"
require "device_specific_commands"
require "device_messages"
require "projector_init"
require "properties"
require "proxy_commands"
require "connections"
require "projector.projector_proxy_class"
require "projector.projector_proxy_commands"
require "projector.projector_proxy_notifies"


-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.driver = "2015.03.31"
end

--[[=============================================================================
    Constants
===============================================================================]]
DRIVER_NAME = Properties["Driver Name"]
PROJECTOR_PROXY_BINDINGID = 5001


--[[=============================================================================
    Initialization Code
===============================================================================]]
function ON_DRIVER_EARLY_INIT.main()
	
end

function ON_DRIVER_INIT.main()
	SetLogName(DRIVER_NAME)

	-- TODO: If cloud based driver then uncomment the following line
	--ConnectURL()
end

function ON_DRIVER_LATEINIT.main()
	C4:urlSetTimeout(20)
end

function ON_DRIVER_EARLY_INIT.projector_driver()

end

function ON_DRIVER_INIT.projector_driver()
    -- TODO: Modify tVolumeRamping to have on entry per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {["state"] = false, ["mode"] = ""}
	
    -- Create an instance of the PROJECTOR Proxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local  bProcessesDeviceMessages = true
    gProjectorProxy = ProjectorProxy:new(PROJECTOR_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping)
    gProjectorProxy:Initialize()
    
    gPowerOnTimer = c4_timer:new("PowerOnTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnTimerExpired)
    gPowerOffTimer = c4_timer:new("PowerOffTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffTimerExpired)
    gPowerOnDelayTimer = c4_timer:new("PowerOnDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnDelayTimerExpired)
    gPowerOffDelayTimer = c4_timer:new("PowerOffDelayTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffDelayTimerExpired)
    gPowerQueryDelayTimer = c4_timer:new("PowerQueryDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerQueryDelayTimerExpired)
    gSetInputDelayTimer = c4_timer:new("SetInputDelayTimer", 2, "SECONDS", OnSetInputDelayTimerExpired, command)   
    
    gLastPowerEvent = ""
    gPowerEventStarted = 0
end

function ON_DRIVER_LATEINIT.projector_driver()

end

--[[=============================================================================
    Driver Code
===============================================================================]]

function QueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end

    gCon:QueueCommand(command, command_delay, delay_units, command_name)	
	
end

function QueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end

    gCon:QueuePriority1Command(command, command_delay, delay_units, command_name)		
	
end

function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x21,0x8901)
		etx = string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x21,0x8901)
		etx = string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueueReferenceCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueReferenceCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueReferenceCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueReferenceCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x3f,0x8901)
		etx = string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x3f,0x8901)
		etx = string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end

	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x21,0x8901)
		etx =  string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x21,0x8901)
		etx =  string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueuePriority1ReferenceCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueuePriority1ReferenceCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueuePriority1ReferenceCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueuePriority1ReferenceCommand()")
	   return
    end

	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x3f,0x8901)
		etx =  string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = string.pack("b>H", 0x3f,0x8901)
		etx =  string.pack('b',0x0A)
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end
